# Copyright (c) 2021 DHI A/S - DHI Water Environment Health 
# All rights reserved.
# 
# This code is licensed under the MIT License.
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files(the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and / or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


import datetime
import pandas as pd
from dateutil import parser

from dhi.platform.base.utils import sanitize_from_to
from dhi.platform.metadata import MetadataClient
from .base import constants
from enum import Enum
from typing import Generator, List, Tuple
from dhi.platform.commonmodels import ItemDefinition, PropertyDefinition, QueryCondition
from .generated.timeseriesgen import TimeSeriesGenClientV2

class TimeseriesDatasetOutput:

    def __init__(
        self,
        id,
        items:Tuple[ItemDefinition],
        timeseries_properties:Tuple[PropertyDefinition],
        metadata:dict={}
    ) -> None:
        self._id = id
        self._items = items
        self._timeseries_properties = timeseries_properties
        self._metadata = metadata
    
    @property
    def id(self):
        return self._id
    
    @property
    def items(self) -> Tuple[ItemDefinition]:
        return self._items

    
    @property
    def timeseries_properties(self) -> Tuple[PropertyDefinition]:
        return self._timeseries_properties

    @property
    def metadata(self) -> dict:
        return self._metadata

    @classmethod
    def from_body(cls, body:dict):
        return cls(
            id = body["id"],
            items = [ItemDefinition.from_body(i) for i in body["items"]],
            timeseries_properties = [PropertyDefinition.from_body(p) for p in body.get("timeSeriesProperties", ())],
            metadata = body.get("metadata", {})
        )

class TimeSeriesDatasetInput:

    def __init__(
        self,
        name:str,
        description:str=None,
        timeseries_schema:Tuple[PropertyDefinition]=(),
        metadata:dict=None,
        properties:dict=None        
    ) -> None:
        self._name = name
        self._description = description
        self._timeseries_schema = timeseries_schema
        self._metadata = metadata
        self._properties = properties

    def body(self) -> dict:
        
        dataset_properties = { "name": self._name }
        
        if self._description is not None:
            dataset_properties["description"] = self._description
        if self._metadata is not None:
            dataset_properties["metadata"] = self._metadata
        if self._properties is not None:
            dataset_properties["properties"] = self._properties
        
        timeseries_properties = [p.body() for p in self._timeseries_schema]
        
        return {
            "timeSeriesSchema": {
                    "properties": timeseries_properties
                },
            "datasetProperties": dataset_properties
        }


class DataFieldDataType(Enum):
    DATETIME = "DateTime"
    SINGLE = "Single"
    DOUBLE = "Double"
    FLAG = "Flag"
    TEXT = "Text"

    @property
    def name(self):
        return self.value

    @classmethod
    def from_string(self, x:str):
        return {
            "datetime": DataFieldDataType.DATETIME,
            "single": DataFieldDataType.SINGLE,
            "double": DataFieldDataType.DOUBLE,
            "flag": DataFieldDataType.FLAG,
            "text": DataFieldDataType.TEXT
        }[x.lower()]


class FlagDefinition:

    def __init__(self, id:int, name:str, level:int, description:str=None) -> None:
        self._id = id
        self._name = name
        self._description = description
        self._level = level
    
    @property
    def id(self) -> str:
        return self._id
    
    @property
    def name(self) -> str:
        return self._name
    
    @property
    def description(self) -> str:
        return self._description
    
    @property
    def level(self) -> int:
        return self._level

    def body(self) -> dict:
        body = {
            "id": self._id,
            "name": self._name,
            "level": self._level
        }

        if self._description is not None:
            body["description"] = self._description

        return body

    @classmethod
    def from_body(cls, body:dict):
        return cls(
            id = body["id"],
            name = body["name"],
            level = body["level"],
            description = body.get("description", None)
        )
    

class DataFieldDefinition:

    def __init__(self, name:str, data_type:DataFieldDataType, flags:Tuple[FlagDefinition]=()) -> None:
        self._name = name
        self._data_type = data_type
        self._flags = flags
    
    @property
    def name(self) -> str: 
        return self._name
    
    @property
    def data_type(self) -> DataFieldDataType: 
        return self._data_type
    
    @property
    def flags(self) -> Tuple[FlagDefinition]: 
        return self._name

    @classmethod
    def from_body(cls, body:dict):
        return cls(
            name = body["name"],
            data_type = DataFieldDataType.from_string(body["dataType"]),
            flags = [FlagDefinition.from_body(f) for f in body.get("flags", ())]
        )

    def body(self) -> dict:
        return {
            "name": self._name,
            "dataType": self._data_type.name,
            "flags": [f.body() for f in self._flags]
        }


class TimeSeriesDefinition:

    def __init__(
        self,
        id:str,
        item:ItemDefinition,
        properties:dict,
        data_fields:Tuple[DataFieldDefinition]=()
    ) -> None:
        self._id = id
        self._item = item
        self._properties = properties
        self._data_fields = data_fields
    
    @property
    def id(self) -> str: 
        return self._id
    
    @property
    def item(self) -> ItemDefinition: 
        return self._item
    
    @property
    def properties(self) -> dict: 
        return self._properties
    
    @property
    def data_fields(self) -> Tuple[DataFieldDefinition]: 
        return self._data_fields

    @classmethod
    def from_body(cls, body:dict):
        return cls(
            id = body.get("id", None),
            item = ItemDefinition.from_body(body["item"]),
            properties = body.get("properties", {}),
            data_fields = [DataFieldDefinition.from_body(f) for f in body.get("dataFields", ())]
        )

    def body(self) -> dict:
        body = {
            "item": self._item.body(),
            "dataFields": [f.body() for f in self._data_fields]
        }

        if self._properties is not None:
            body["properties"] = self._properties

        if self._id:
            body["id"] = self._id
        
        return body


class TimeSeriesClientV2(TimeSeriesGenClientV2):
    def __init__(self, inspectFnc=TimeSeriesGenClientV2.DefaultInspectFnc, **kwargs):
        super().__init__(inspectFnc, **kwargs)

class TimeSeriesClient():
    
    def __init__(self, inspectFnc=TimeSeriesGenClientV2.DefaultInspectFnc, **kwargs):
        self._metadata = kwargs.get("MetadataClient", MetadataClient(inspectFnc, **kwargs))
        self._ts2 = kwargs.get("TimeSeriesGenClientV2", TimeSeriesGenClientV2(inspectFnc, **kwargs))

    def _read_timeseries_data_from_response_data(self, data, data_fields:Tuple[DataFieldDefinition]=()):
        date_times = []
        values = []
        field_values = {}
        for field in data_fields:
            field_values[field.name] = []
        
        for row in data:
            s = row[0]
            t = parser.parse(s)
            date_times.append(t)
            v = row[1]
            values.append(None if v is None else v)
            for f, field in enumerate(data_fields):
                field_value = row[2+f]
                field_values[field.name].append(field_value)
        
        field_values.update({"values": values})
        timeseries_data = pd.DataFrame(field_values, index=date_times)

        return timeseries_data

    def get_timeseries_dataset(self, project_id, dataset_id) -> TimeseriesDatasetOutput:
        """
        Get details about time series dataset
        :param project_id: ID of the project with the timeseries dataset
        :param dataset_id: ID of the timeseries dataset
        :return: Time series dataset
        :rtype: TimeseriesDatasetOutput
        """
        response = self._ts2.GetTimeSeriesDatasetV2(project_id, dataset_id)
        return TimeseriesDatasetOutput.from_body(response.Body)
    
    def create_timeseries_dataset(self, project_id, input:TimeSeriesDatasetInput):
        """
        Create a new time series dataset

        :param project_id: ID of the project with the timeseries dataset
        :param input: Timeseries dataset input
        :return: Timeseries dataset
        :rtype: TimeseriesDatasetOutput
        """
        response = self._ts2.CreateTimeseriesDatasetV2(project_id, input.body())
        output = TimeseriesDatasetOutput.from_body(response.Body)
        self._metadata.wait_until_dataset_exists(output.id)
        return output

    def create_timeseries_dataset_from_schema(self, project_id, name:str, description:str=None, timeseries_schema:Tuple[PropertyDefinition]=()):
        """
        Create a new timeseries dataset
        
        :param project_id: ID of the project with the timeseries dataset
        :param name: Dataset name
        :param description: Dataset description
        :param timeseries_schema: Definition of the properties each time series should have
        :return: Timeseries dataset
        :rtype: TimeSeriesDatasetOutput
        """
        input = TimeSeriesDatasetInput(name, description, timeseries_schema)
        response = self._ts2.CreateTimeseriesDatasetV2(project_id, input.body())
        output =  TimeseriesDatasetOutput.from_body(response.Body)
        self._metadata.wait_until_dataset_exists(output.id)
        return output
    
    def add_timeseries_dataset_properties(self, project_id, properties:Tuple[PropertyDefinition]) -> bool:
        """
        Add new property definitions into time series dataset.
        
        :param project_id: ID of the project with the timeseries dataset
        :param properties: Property definitions
        :return: True if the dataset was created successfully
        :rtype: bool
        """
        input = {"properties": [p.body() for p in properties]}
        response = self._ts2.AddDatasetSchemaPropertiesV2(project_id, input)
        return response.IsOk

    def get_timeseries(self, project_id, dataset_id, timeseries_id) -> TimeSeriesDefinition:
        """
        Get details about time series
        
        :param project_id: ID of the project with the timeseries dataset
        :param dataset_id: ID of the timeseries dataset
        :param timeseries_id: ID of the timeseries to get
        :return: Timeseries definition
        :rtype: TimeSeriesDefinition
        """
        response = self._ts2.GetTimeSeriesByIdV2(project_id, dataset_id, timeseries_id)
        return TimeSeriesDefinition.from_body(response.Body)

    def get_timeseries_values(self, project_id, dataset_id, timeseries_id, from_:datetime.datetime=None, to:datetime.datetime=None) -> pd.DataFrame:
        """
        Get time series values
        
        :param project_id: ID of the project with the timeseries dataset
        :param dataset_id: ID of the timeseries dataset
        :param timeseries_id: ID of the timeseries to get values from
        :return: Pandas DataFrame
        """
        from_, to = sanitize_from_to(from_, to)
        ts = self.get_timeseries(project_id, dataset_id, timeseries_id)
        response = self._ts2.GetTimeSeriesDataV2(project_id, dataset_id, timeseries_id, from_, to)
        data = response.Body["data"]
        return self._read_timeseries_data_from_response_data(data, ts.data_fields)
    
    def list_timeseries(self, project_id, dataset_id) -> Generator[TimeSeriesDefinition, None, None]:
        """
        Get list of dataset's time series
        
        :param project_id: ID of the project with the timeseries dataset
        :param dataset_id: ID of the timeseries dataset
        :return: Generator of timeseries definitions
        """
        response = self._ts2.GetAllTimeSeriesV2(project_id, dataset_id)
        data = response.Body["data"]
        return (TimeSeriesDefinition.from_body(t) for t in data)

    def query_timeseries(self, project_id, dataset_id, conditions:Tuple[QueryCondition]) -> Generator[TimeSeriesDefinition, None, None]:
        """
        Query time series by name or property value(s)
        
        Query filter example to get "Rainfall" time series with "CX" property equal to {1}: 
        conditions = (  AttributeQueryCondition("Item", "Equal", "Rainfall"),
                        "AttributeQueryCondition("CX", "Equal", 1),
                    )

        :param project_id: ID of the project with the timeseries dataset
        :param dataset_id: ID of the timeseries dataset
        :param conditions: QueryConditions the timeseries should fulfill.
        :return: Generator of timeseries definitions
        """
        input = { "conditions": [c.body() for c in conditions] }
        response = self._ts2.GetQueryTimeSeriesV2(project_id, input, dataset_id)
        return (TimeSeriesDefinition.from_body(t) for t in response.Body["data"])
    
    def get_multiple_timeseries_values(self, project_id, dataset_id, timeseries_ids:Tuple[str], from_=None, to=None) -> Generator[pd.DataFrame, None, None]:
        """
        Get timeseries values

        :param project_id: ID of the project with the timeseries dataset
        :param dataset_id: ID of the timeseries dataset
        :param timeseries_id: ID of the timeseries
        :param timeseries_ids: IDs of the timeseries to get values from
        :return: Generator of data frames where each data frame corresponds to a timeseries id
        """
        from_, to = sanitize_from_to(from_, to)

        data_fields = [self.get_timeseries(project_id, dataset_id, i).data_fields for i in timeseries_ids]
        response = self._ts2.GetMultiTimeSeriesDataV2(project_id, timeseries_ids, dataset_id, from_, to)
        data = response.Body["data"]

        for i, d in enumerate(data):
            yield self._read_timeseries_data_from_response_data(d, data_fields[i])
        
    def add_timeseries(self, project_id, dataset_id, item:ItemDefinition, properties:dict=None, data_fields:Tuple[DataFieldDefinition]=()):
        """
        Create a new timeseries

        :param project_id: ID of the project with the timeseries dataset
        :param dataset_id: ID of the timeseries dataset
        :param item: Definition of the timeseries item
        :param properties: Name-value dictionary of timeseries properties
        :param data_fields: Definition of the data fields of the timeseries
        :return: Timeseries definition
        :rtype: TimeSeriesDefinition
        """
        return self.add_timeseries_with_id(project_id, dataset_id, None, item, properties, data_fields)     

    def add_timeseries_with_id(self, project_id, dataset_id, timeseries_id:str, item:ItemDefinition, properties:dict=None, data_fields:Tuple[DataFieldDefinition]=()):
        """
        Create a new timeseries

        :param project_id: ID of the project with the timeseries dataset
        :param dataset_id: ID of the timeseries dataset
        :param timeseries_id: Desired ID of the timeseries
        :param item: Definition of the timeseries item
        :param properties: Name-value dictionary of timeseries properties
        :param data_fields: Definition of the data fields of the timeseries
        :return: Timeseries definition
        :rtype: TimeSeriesDefinition
        """
        input = TimeSeriesDefinition(
            timeseries_id,
            item,
            properties,
            data_fields
        )
        
        response = self._ts2.AddTimeSeriesV2(project_id, input.body(), dataset_id)
        return TimeSeriesDefinition.from_body(response.Body)
    
    def update_timeseries_properties(self, project_id, dataset_id, timeseries_id, properties:dict) -> bool:
        """
        Update all timeseries properties

        :param project_id: ID of the project with the timeseries dataset
        :param dataset_id: ID of the timeseries dataset
        :param timeseries_id: ID of the timeseries
        :param properties: Name-value dictionary of the new properties for the timeseries.
            Note that properties not included in the dictionary will be set to None.
            Always include full set of properties as you want it to look after the update
        :return: True if the timeseries properties were updated successfully
        :rtype: bool
        """
        input = { "properties": properties }
        response = self._ts2.SetAllTimeSeriesPropertiesV2(project_id, input, dataset_id, timeseries_id)
        return response.IsOk

    def delete_timeseries(self, project_id, dataset_id, timeseries_id) -> bool:
        """
        Delete timeseries from dataset

        :param project_id: ID of the project with the timeseries dataset
        :param dataset_id: ID of the timeseries dataset
        :param timeseries_id: ID of the timeseries
        :return: True if the timeseries was removed successfully
        :rtype: bool
        """
        response = self._ts2.DeleteTimeSeriesV2(project_id, dataset_id, timeseries_id)
        return response.IsOk

    def update_timeseries_property(self, project_id, dataset_id, timeseries_id, property_name, property_value) -> bool:
        """
        Update a single timeseries property

        :param project_id: ID of the project with the timeseries dataset
        :param dataset_id: ID of the timeseries dataset
        :param timeseries_id: ID of the timeseries
        :param property_name: Name of the property to update
        :param property_value: New value for the property to update
        :return: True if the value was updated successfully
        :rtype: bool
        """
        input = { "property": property_name, "value": property_value }
        response = self._ts2.SetTimeSeriesPropertyV2(project_id, input, dataset_id, timeseries_id)
        return response.IsOk

    def delete_timeseries_values(self, project_id, dataset_id, timeseries_id, from_:datetime.datetime=None, to:datetime.datetime=None) -> bool:
        """
        Delete timeseries values

        :param project_id: ID of the project with the timeseries dataset
        :param dataset_id: ID of the timeseries dataset
        :param timeseries_id: ID of the timeseries
        :param from_: optional time from which the values should be removed
        :param to: optional time to which the values should be removed
        :return: True if the values were removed successfully
        :rtype: bool
        """
        from_, to = sanitize_from_to(from_, to)

        response = self._ts2.DeleteTimeSeriesValuesV2(project_id, dataset_id, timeseries_id, from_, to)
        return response.IsOk

    def add_timeseries_values(self, project_id, dataset_id, timeseries_id, data:pd.DataFrame,
        datetimeindex=0, valueindex=1, flagindex=None
    ) -> bool:
        """
        Add timeseries values to a timeseries
        
        :param data: Pandas data frame with index of datetime.datetime and one column matching the timeseries data type
        :param datetimeindex: Optional parameter indicating the datetime field index in the data frame, default is 0
        :param valueindex: Optional parameter indicating the value field index in the data frame, default is 1
        :param flagindex: Optional, if the data frame contains a flag, provide the flag field index using this parameter, default is None
        :return: A value indicating whether the operation was successfull (True) or not (False)
        :rtype: bool
        """
        separator = ";"
        input = data.to_csv(sep=separator, date_format=constants.DATETIMEFORMAT_SECONDS, header=False)
        response = self._ts2.UploadCsvDataV2(project_id, input, dataset_id, timeseries_id, datetimeindex=datetimeindex, delimiter=separator, decimalseparator=".", valueindex=valueindex, flagindex=flagindex, content_type="text/plain")
        return response.IsOk


if __name__ == '__main__':
    print(__file__)
    print(dir())



# Copyright (c) 2021 DHI A/S - DHI Water Environment Health 
# All rights reserved.
# 
# This code is licensed under the MIT License.
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files(the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and / or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import itertools
from .base.client import Contracts


class RawClientV2Contracts(Contracts):
    @classmethod
    def PrepareUploadBulkFilesInput(cls, input, id=None, sastoken=None, name=None, url=None, lastmodified=None, size=None, forcecopy=None):
        body = input.copy() if input else {}
        datasets = body.get("datasets")
        if not datasets:
            datasets = []
            body["datasets"] = datasets
        allfields = itertools.zip_longest(id if id else [], sastoken if sastoken else [], name if name else [], url if url else [], lastmodified if lastmodified else [], size if size else [], forcecopy if forcecopy else [])
        for i, sas, nm, u, lm, sz, fc in allfields:
            item = {}
            cls.SetBodyField(item, "id", i)
            cls.SetBodyField(item, "sastoken", sas)
            cls.SetBodyField(item, "name", nm)
            cls.SetBodyField(item, "url", u)
            cls.SetBodyField(item, "lastmodified", lm)
            cls.SetBodyField(item, "size", sz)
            cls.SetBodyField(item, "forcecopy", fc)
            datasets.append(item)
        return body

    @classmethod
    def PrepareGetCopyFileStatusInput(cls, input, id=None, copyoperationid=None):
        body = input.copy() if input else {}
        datasets = body.get("datasets")
        if not datasets:
            datasets = []
            body["datasets"] = datasets
        allfields = itertools.zip_longest(id if id else [], copyoperationid if copyoperationid else [])
        for i, cpi in allfields:
            item = {}
            cls.SetBodyField(item, "id", i)
            cls.SetBodyField(item, "copyOperationId", cpi)
            datasets.append(item)
        return body


if __name__ == '__main__':
    print(__file__)
    print(dir())



# Copyright (c) 2021 DHI A/S - DHI Water Environment Health 
# All rights reserved.
# 
# This code is licensed under the MIT License.
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files(the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and / or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import os
import time
import uuid
import datetime
from abc import ABC, abstractmethod
from enum import Enum
from typing import Generator, List, Tuple
from dhi.platform.base.exceptions import MikeCloudException
from dhi.platform.raw import RawClientV2
from dhi.platform.transferuploadhelper import TransferUploadHelper
from .generated.metadatagen import MetadataGenClientV1
from .metadata import MetadataClientV2, MetadataClientV3
from .base import constants
from azure.storage.blob import BlobClient
from .commonmodels import PropertyDefinition, ItemId, TemporalValueFilter, UnitId, SpatialFilter, TemporalFilter, VerticalFilter


class ParameterDefinition:
    def __init__(self, name:str, description:str, data_type:str, required:bool, default:object=None, allowed:Tuple[object]=None) -> None:
        self._name = name
        self._description = description
        self._data_type = data_type
        self._required = required
        self._default = default
        self._allowed = allowed

    @property
    def name(self):
        return self._name

    @property
    def description(self):
        return self._description

    @property
    def data_type(self):
        return self._data_type

    @property
    def required(self):
        return self._required

    @property
    def default(self):
        return self._default

    @property
    def allowed(self):
        return self._allowed

    @classmethod
    def from_body(cls, body:dict):
        return cls(
            name = body["name"],
            description = body.get("description", None),
            data_type = body["dataType"],
            required = bool(body.get("required", False)),
            default = body.get("defaultValue", None),
            allowed = body.get("allowedValues", None),
        )


class ConverterOutput:
    def __init__(self, name, description, format, parameters) -> None:
        self._name = name,
        self._description = description
        self._format = format
        self._parameters = parameters
    
    @property
    def name(self):
        return self._name

    @property
    def description(self):
        return self._description

    @property
    def format(self):
        return self._format

    @property
    def parameters(self):
        return self._parameters

    @classmethod
    def from_body(cls, body:dict):
        return cls(
            name = body["name"],
            description = body.get("description", None),
            format = body.get("datasetFormat"),
            parameters = [ParameterDefinition.from_body(p) for p in body.get("parameters", [])]
        )


class ConverterFilter(Enum):
    ALL = 0
    FILE = 1
    DEDICATED = 2


class TransferType(Enum):
    IMPORT = 0
    EXPORT = 1
    CONVERSION = 2
    APPEND = 3
    UPDATE = 4

    @classmethod
    def from_string(self, x:str):
        return {
            "import": TransferType.IMPORT,
            "export": TransferType.EXPORT,
            "conversion": TransferType.CONVERSION,
            "append": TransferType.APPEND,
            "update": TransferType.UPDATE,
        }[x.lower()]


class ImportDestination(Enum):
    DEDICATED = 0
    PROJECT = 1

    @classmethod
    def from_string(self, x:str):
        return {
            'dedicated': ImportDestination.DEDICATED, 
            'project': ImportDestination.PROJECT
        }.get(x.lower(), ImportDestination.PROJECT)


class TransferStatus(Enum):
    NONE = 0
    PENDING = 1
    INPROGRESS = 2
    COMPLETED = 3
    ERROR = 4

    @property
    def name(self):
        return "InProgress" if self is TransferStatus.INPROGRESS else super(TransferStatus, self).name.title()

    @classmethod
    def from_string(self, x:str):
        return {
            "none": TransferStatus.NONE,
            "pending": TransferStatus.PENDING,
            "inprogress": TransferStatus.INPROGRESS,
            "completed": TransferStatus.COMPLETED,
            "error": TransferStatus.ERROR,
        }.get(x.lower(), TransferStatus.NONE)

    
class AggregationType(Enum):
    MAX = 0
    MIN = 1
    AVG = 2

class CoordinateInterpretationType(Enum):
    CENTER_OF_CELL = 0
    BOTTOM_LEFT_BORDER = 1
    BORDERS = 2

    @property
    def name(self):
        return {
            CoordinateInterpretationType.CENTER_OF_CELL: "CenterOfCell",
            CoordinateInterpretationType.BOTTOM_LEFT_BORDER: "BottomLeftBorder",
            CoordinateInterpretationType.BORDERS: "Borders"
        }[self]

    @classmethod
    def from_string(self, x:str):
        return {
            "centerofcell": CoordinateInterpretationType.CENTER_OF_CELL,
            "bottomleftborder": CoordinateInterpretationType.BOTTOM_LEFT_BORDER,
            "borders": CoordinateInterpretationType.BORDERS,
        }.get(x.lower(), CoordinateInterpretationType.CENTER_OF_CELL)


class ImportParameters:
    def __init__(
        self,
        append_dataset_id,
        upload_url:str,
        file_name:str,
        srid:int,
        arguments:dict,
        destinations:Tuple[ImportDestination]
    ) -> None:
        self._append_dataset_id = append_dataset_id
        self._upload_url = upload_url
        self._file_name = file_name
        self._srid = srid
        self._arguments = arguments
        self._destinations = destinations

    @property
    def append_dataset_id(self):
        return self._append_dataset_id

    @property
    def upload_url(self):
        return self._upload_url

    @property
    def file_name(self):
        return self._file_name

    @property
    def srid(self):
        return self._srid

    @property
    def arguments(self):
        return self._arguments

    @property
    def destinations(self):
        return self._destinations
    
    @classmethod
    def from_body(cls, body:dict):
        return cls(
            append_dataset_id = body.get("appendDatasetId"),
            upload_url = body.get("uploadUrl"),
            file_name = body.get("fileName"),
            srid = body.get("srid"),
            arguments = body.get("arguments", {}),
            destinations = body.get("destinations", [])
        )


class ExportParameters:
    def __init__(
        self,
        dataset_id,
        output_file_name,
        srid:int,
        arguments:dict
    ) -> None:
        self._dataset_id = dataset_id
        self._output_file_name = output_file_name
        self._srid = srid
        self._arguments = arguments
    
    @property
    def dataset_id(self):
        return self._dataset_id
    
    @property
    def output_file_name(self):
        return self._output_file_name
    
    @property
    def srid(self):
        return self._srid
    
    @property
    def arguments(self):
        return self._arguments

    @classmethod
    def from_body(cls, body:dict):
        return cls(
            dataset_id = body.get("datasetId"),
            output_file_name = body.get("outputFileName"),
            srid = body.get("srid"),
            arguments = body.get("arguments", {})
        )

class DatasetTransferInput:
    def __init__(self, name:str, description:str=None, metadata:dict=None, properties:dict=None) -> None:
        self._name = name
        self._description = description
        self._metadata = metadata
        self._properties = properties
    
    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, value):
        self._description = value

    @property
    def metadata(self):
        return self._metadata

    @property
    def properties(self):
        return self._properties

    @classmethod
    def from_body(cls, body:dict):
        return cls(
            name = body["name"],
            description = body.get("description"),
            metadata = body.get("metadata", {}),
            properties = body.get("properties", {})
        )
    
    def body(self):
        body = { "name": self._name }

        if self._description:
            body["description"] = self._description

        if self._metadata is not None:
            body["metadata"] = self._metadata

        if self._properties is not None:
            body["properties"] = self._properties

        return body


class ImportResult:
    def __init__(self, project_id, dataset_id) -> None:
        self._project_id = project_id
        self._dataset_id = dataset_id
    
    @property
    def project_id(self):
          return self._project_id
    
    @property
    def dataset_id(self):
          return self._dataset_id

    @classmethod
    def from_body(cls, body:dict):
        return cls(
            project_id = body.get("projectId"),
            dataset_id = body.get("datasetId")
        )

        
def get_datetime_from_body(property_name:str, body:dict):
    if property_name not in body:
        return None
    else:
        property = body[property_name]
        count = len(property)
        if count == 19:
            return datetime.datetime.strptime(property, constants.DATETIMEFORMAT_SECONDS)
        else:
            return datetime.datetime.strptime(property[:26], constants.DATETIMEFORMAT)

class BaseEntityOutput:
    def __init__(
        self, 
        id,
        created_at,
        created_by,
        updated_at,
        updated_by,
        deleted_at,
        deleted_by
    ) -> None:
        self._id = id
        self._created_at = created_at
        self._created_by = created_by
        self._updated_at = updated_at
        self._updated_by = updated_by
        self._deleted_at = deleted_at
        self._deleted_by = deleted_by
    
    @property
    def id(self):
        return self._id
    
    @property
    def created_at(self):
        return self._created_at
    
    @property
    def created_by(self):
        return self._created_by
    
    @property
    def updated_at(self):
        return self._updated_at
    
    @property
    def updated_by(self):
        return self._updated_by
    
    @property
    def deleted_at(self):
        return self._deleted_at
    
    @property
    def deleted_by(self):
        return self._deleted_by

    @classmethod
    def from_body(cls, body:dict):
        return cls(
            id = body["id"],
            created_at = get_datetime_from_body("createdAt", body),
            created_by = get_datetime_from_body("createdAy", body),
            updated_at = get_datetime_from_body("updatedAt", body),
            updated_by = get_datetime_from_body("updatedAy", body),
            deleted_at = get_datetime_from_body("deletedAt", body),
            deleted_by = get_datetime_from_body("deletedAy", body)
        )
        

class TransferOutput(BaseEntityOutput):
    def __init__(
        self,
        id,
        created_at,
        created_by,
        updated_at,
        updated_by,
        deleted_at,
        deleted_by,
        transfer_type,
        status,
        format,
        project_id,
        import_parameters,
        export_parameters,
        dataset_import_data,
        download_path,
        error_message,
        import_results
    ) -> None:
        super(TransferOutput, self).__init__(id, created_at, created_by, updated_at, updated_by, deleted_at, deleted_by)
        self._transfer_type = transfer_type
        self._status = status
        self._format = format
        self._project_id = project_id
        self._import_parameters = import_parameters
        self._export_parameters = export_parameters
        self._dataset_import_data = dataset_import_data
        self._download_path = download_path
        self._error_message = error_message
        self._import_results = import_results
    
    @property
    def transfer_type(self):
        return self._transfer_type
    
    @property
    def status(self):
        return self._status
    
    @property
    def format(self):
        return self._format
    
    @property
    def project_id(self):
        return self._project_id
    
    @property
    def import_parameters(self):
        return self._import_parameters
    
    @property
    def export_parameters(self):
        return self._export_parameters
    
    @property
    def dataset_import_data(self):
        return self._dataset_import_data
    
    @property
    def download_path(self):
        return self._download_path
    
    @property
    def error_message(self):
        return self._error_message
    
    @property
    def import_results(self) -> List[ImportResult]:
        return self._import_results

    @classmethod
    def from_body(cls, body: dict):

        import_parameters_body = body.get("importParameters")
        import_parameters = ImportParameters.from_body(import_parameters_body) if import_parameters_body else None
        export_parameters_body = body.get("exportParameters")
        export_parameters = ExportParameters.from_body(export_parameters_body) if export_parameters_body else None
        dataset_import_data_body = body.get("datasetImportData")
        dataset_import_data = DatasetTransferInput.from_body(dataset_import_data_body) if dataset_import_data_body else None
        import_results = [ImportResult.from_body(r) for r in body.get("importResults", [])]

        return cls(
            id = body.get("id"),
            created_at = body.get("createdAt"),
            created_by = body.get("createdBy"),
            updated_at = body.get("updatedAt"),
            updated_by = body.get("updatedBy"),
            deleted_at = body.get("deletedAt"),
            deleted_by = body.get("deletedBy"),
            transfer_type = TransferType.from_string(body.get("type")),
            status = TransferStatus.from_string(body.get("status")),
            format = body.get("format"),
            project_id = body.get("projectId"),
            import_parameters = import_parameters,
            export_parameters = export_parameters,
            dataset_import_data = dataset_import_data,
            download_path = body.get("downloadPath", None),
            error_message = body.get("errorMessage", None),
            import_results = import_results
        )


class SucceededUploadOutput:
    def __init__(self, file_name, dataset_id) -> None:
        self._file_name = file_name
        self._dataset_id = dataset_id
    
    @property
    def file_name(self):
        return self._file_name
    
    @property
    def dataset_id(self):
        return self._dataset_id
    
    @classmethod
    def from_body(cls, body:dict):
        return cls(
            file_name = body["fileName"],
            dataset_id = body["datasetId"]
        )


class FailedUploadOutput:
    def __init__(self, file_name, message) -> None:
        self._file_name = file_name
        self._messsage = message
    
    @property
    def file_name(self): 
        return self._file_name
    
    @property
    def message(self):
        return self._messsage

    @classmethod
    def from_body(cls, body:dict):
        return cls(
            file_name = body["fileName"],
            message = body.get("message")
        )


class StagedFileUploadInput:
    def __init__(self, url:str, file_name:str) -> None:
        self._url = url
        self._file_name = file_name
    
    @property
    def url(self):
        return self._url
    
    @property
    def file_name(self):
        return self._file_name
    
    def body(self):
        return {
            "url": self._url,
            "fileName": self._file_name
        }


class StagedFilesUploadInput:
    def __init__(self, files:Tuple[StagedFileUploadInput], destination_path:str=None, create_destination_path_if_not_exists:bool=True) -> None:
        self._files = files
        self._destination_path = destination_path
        self._create_destination_path_if_not_exists = create_destination_path_if_not_exists
    
    
    @property
    def files(self):
            return self._files
    
    @property
    def destination_path(self): 
            return self._destination_path
    
    @property
    def create_destination_path_if_not_exists(self): 
            return self._create_destination_path_if_not_exists

    def body(self):
        body = {
            "files": [s.body() for s in self._files],
            "createDestinationPathIfNotExists": self._create_destination_path_if_not_exists
        }
        
        if self._destination_path:
            body["destingationPath"] = self._destination_path
            
        return body

class StagedFilesUploadOutput:
    def __init__(self, datasets:Tuple[SucceededUploadOutput], failures:Tuple[FailedUploadOutput]) -> None:
        self._datasets = datasets
        self._failures = failures

    @property
    def datasets(self):
        return self._datasets
    
    @property
    def failures(self):
        return self._failures

    @classmethod
    def from_body(cls, body:dict):
        return cls(
            datasets = [SucceededUploadOutput.from_body(d) for d in body["datasets"]],
            failures = [FailedUploadOutput.from_body(d) for d in body["failures"]]
        )

class TransferSummaryOutput:
    def __init__(
        self,
        id, 
        created_at,
        created_by,
        transfer_type:TransferType,
        format,
        status:TransferStatus
    ) -> None:
        self._id = id
        self._created_at = created_at
        self._created_by = created_by
        self._transfer_type = transfer_type
        self._format = format
        self._status = status
    
    @property
    def id(self):
        return self._id
    
    @property
    def created_at(self):
        return self._created_at
    
    @property
    def created_by(self):
        return self._created_by
    
    @property
    def transfer_type(self):
        return self._transfer_type
    
    @property
    def format(self):
        return self._format
    
    @property
    def status(self):
        return self._status

    @classmethod
    def from_body(cls, body:dict):
        return cls(
            id = body["id"], 
            created_at = get_datetime_from_body("createdAt", body),
            created_by = body.get("createdBy", None),
            transfer_type = TransferType.from_string(body["type"]),
            format = body.get("format"),
            status = TransferStatus.from_string(body.get("status"))
        )


class TransformationBase:
    pass

    @abstractmethod
    def body(self) -> dict:
        return {}

class CrsTransformation(TransformationBase):
    def __init__(self, output_srid:int, override_input_srid:int=None) -> None:
        super().__init__()
        self._output_srid = output_srid
        self._override_input_srid = override_input_srid

    @property
    def output_srid(self):
        return self._output_srid

    @property
    def override_input_srid(self):
        return self._override_input_srid
    
    def body(self) -> dict:
        body = { 
            "type": "CrsTransformation",
            "outputSrid": self._output_srid 
        }
        if self._override_input_srid:
            body["inputSrid": self._override_input_srid]
        return body
        

class ItemsFilter:
    pass

    @abstractmethod
    def body(self):
        pass


class ItemIndexFilter(ItemsFilter):
    def __init__(self, item_indices:Tuple[int]) -> None:
        super().__init__()
        self._item_indices = item_indices
    
    def body(self) -> dict:
        return { "itemIndices": list(self._item_indices) }
    

class ItemNameFilter(ItemsFilter):
    def __init__(self, names:Tuple[str]) -> None:
        super().__init__()
        self._names = names
    
    def body(self) -> dict:
        return { "names": list(self._names) }


class Aggregation:
    def __init__(self, items_filter:ItemsFilter, aggregation_type:AggregationType) -> None:
        self._items_filter = items_filter
        self._aggregation_type = aggregation_type
    
    def body(self) -> dict:
        return {
            "itemsFilter": self._items_filter.body(),
            "aggregationType": self._aggregation_type.name.title()
        }


class AggregationTransformation(TransformationBase):
    def __init__(self, aggregations:Tuple[Aggregation]) -> None:
        super().__init__()
        self._aggregations = aggregations

    def body(self) -> dict:
        return {
            "type": "AggregationTransformation",
            "aggregations": [a.body() for a in self._aggregations] 
        }


class CsScriptValueTransformation(TransformationBase):
    """Not implemented in Python SDK"""

    def __init__(self) -> None:
        super().__init__()
        raise NotImplementedError("CsScriptValueTransformation is not implemented in the Python SDK")


class ItemFilterTransformation(TransformationBase):
    def __init__(self, item_filter:Tuple[ItemsFilter]) -> None:
        super().__init__()
        self._item_filter = item_filter

    def body(self) -> dict:
        return { 
            "type": "ItemFilterTransformation",
            "itemFilter": [i.body() for i in self._item_filter] 
        }


class ItemRedefinition:
    def __init__(self, original_name:str, new_name:str, new_item_id:ItemId, new_unit_id:UnitId) -> None:
        self._original_name = original_name
        self._new_name = new_name
        self._new_item_id = new_item_id
        self._new_unit_id = new_unit_id

    def body(self) -> dict:
        body = {
            "originalName": self._original_name,
            "newName": self._new_name,
            "newItemId": self._new_item_id,
            "newUnitId": self._new_unit_id            
        }
        return body


class ItemTransformation(TransformationBase):
    def __init__(self, item_redefinitions:Tuple[ItemRedefinition]) -> None:
        super().__init__()
        self._item_redefinitions = item_redefinitions

    def body(self) -> dict:
        body = {
            "type": "ItemTransformation",
            "itemRedefinitions": [r.body() for r in self._item_redefinitions]
        }
        return body

class SpatialFilterTransformation(TransformationBase):
    def __init__(self, spatial_filter:SpatialFilter) -> None:
        super().__init__()
        self._spatial_filter = spatial_filter
    
    def body(self) -> dict:
        body = {
            "type": "SpatialFilterTransformation",
            "spatialFilter": self._spatial_filter.body()
        }
        return body


class TemporalFilterTransformation(TransformationBase):
    def __init__(self, temporal_filter:TemporalFilter) -> None:
        super().__init__()
        self._temporal_filter = temporal_filter
    
    def body(self) -> dict:
        body = {
            "type": "TemporalFilterTransformation",
            "temporalFilter": self._temporal_filter.body()
        }
        return body
        

class VerticalFilterTransformation(TransformationBase):
    def __init__(self, vertical_filter:VerticalFilter) -> None:
        super().__init__()
        self._vertical_filter = vertical_filter
    
    def body(self) -> dict:
        body = {
            "type": "VerticalFilterTransformation",
            "verticalFilter": self._vertical_filter.body()
        }
        return body


class TransferWriterBase:
    def __init__(self, parameters:tuple = ()) -> None:
        self._parameters = list(parameters)

    def __repr__(self) -> str:
        return "Writer " + str(getattr(self, "name", None))
    
    def add_parameter(self, parameter):
        self._parameters.append(parameter)
    
    @property
    def parameters(self):
        return self._parameters

    def _with_timeseries_properties_impl(self, properties:Tuple[PropertyDefinition]):
        self._parameters.append((
                "TimeSeriesProperties",
                [p.body() for p in properties]
        ))

    def _with_write_as_float_impl(self):
        self._parameters(("WriteAsFloat", True))


class GenericWriter(TransferWriterBase):
    def __init__(self, name:str, parameters: tuple = ()) -> None:
        super().__init__(parameters=parameters)
        self._name = name
    
    @property
    def name(self):
        return self._name
    
    def with_timeseries_properties(self, properties: tuple):
        return super()._with_timeseries_properties_impl(properties)
    
    def with_write_as_float(self):
        return super()._with_write_as_float_impl()


class MDWriter(TransferWriterBase):
    def __init__(self, parameters:Tuple = ()) -> None:
        super().__init__(parameters=parameters)
    
    @property
    def name(self):
        return "MDWriter"


class Dfs2Writer(TransferWriterBase):
    def __init__(self, parameters:Tuple = ()) -> None:
        super().__init__(parameters=parameters)
    
    @property
    def name(self):
        return "Dfs2Writer"

    def with_write_as_float(self):
        return super()._with_write_as_float_impl()


class TSWriter(TransferWriterBase):
    def __init__(self, parameters:Tuple = ()) -> None:
        super().__init__(parameters=parameters)
    
    @property
    def name(self):
        return "TSWriter"

    def with_timeseries_properties(self, properties: tuple):
        return super()._with_timeseries_properties_impl(properties)


class FileWriter(TransferWriterBase):
    def __init__(self) -> None:
        super().__init__(parameters=())
    
    @property
    def name(self):
        return "FileWriter"


class TransferReaderBase:
    def __init__(self, parameters:tuple = ()) -> None:
        self._parameters = list(parameters)

    def __repr__(self) -> str:
        return "Reader " + str(getattr(self, "name", None))

    def add_parameter(self, parameter):
        self._parameters.append(parameter)
    
    @property
    def parameters(self):
        return self._parameters

    def _with_srid_impl(self, srid:int):
        self._parameters.append(('SRID', srid))
    
    def _with_coordinate_interpretation_type_impl(self, coordinate_interpretation_type:CoordinateInterpretationType):
        self._parameters.append(('CoordinateInterpretationType', coordinate_interpretation_type.name))

    def _with_allowed_items_impl(self, allowed_items:Tuple[str]):
        self._parameters.append(("AllowedItemNames", list(allowed_items)))


class GenericReader(TransferReaderBase):
    def __init__(self, name:str, parameters: tuple = ()) -> None:
        super().__init__(parameters=parameters)
        self._name = name
    
    @property
    def name(self):
        return self._name
    
    def with_srid(self, srid:int):
        super()._with_srid_impl(srid)
    
    def with_coordinate_interpretation_type(self, coordinate_interpretation_type: CoordinateInterpretationType):
        return super()._with_coordinate_interpretation_type_impl(coordinate_interpretation_type)
    
    def with_allowed_items(self, allowed_items: Tuple[str]):
        return super()._with_allowed_items_impl(allowed_items)


class MDReader(TransferReaderBase):
    def __init__(self, parameters:Tuple = ()) -> None:
        super().__init__(parameters=parameters)
    
    @property
    def name(self):
        return "MDReader"


class DfsuReader(TransferReaderBase):
    def __init__(self, parameters: Tuple = ()) -> None:
        super().__init__(parameters=parameters)
    
    @property
    def name(self):
        return "DfsuReader"
    
    def with_srid(self, srid: int):
        return super()._with_srid_impl(srid)

    def with_allowed_items(self, allowed_items: Tuple[str]):
        return super()._with_allowed_items_impl(allowed_items)   


class Dfs2Reader(TransferReaderBase):
    def __init__(self, parameters: Tuple = ()) -> None:
        super().__init__(parameters=parameters)
    
    @property
    def name(self):
        return "Dfs2Reader"
    
    def with_allowed_items(self, allowed_items: Tuple[str]):
        return super()._with_allowed_items_impl(allowed_items)

    def with_coordinate_interpretation_type(self, coordinate_interpretation_type: CoordinateInterpretationType):
        return super()._with_coordinate_interpretation_type_impl(coordinate_interpretation_type)

    def with_srid(self, srid: int):
        return super()._with_srid_impl(srid)


class FileReader(TransferReaderBase):
    def __init__(self) -> None:
        super().__init__(parameters=())
    
    @property
    def name(self):
        return "FileReader"


class TransferPipeline:
    def __init__(self, reader, writer, transformations=[]) -> None:
        self._reader = reader
        self._writer = writer
        self._transformations = transformations
    
    @property
    def reader(self):
        return self._reader
    
    @property
    def writer(self):
        return self._writer
    
    @property
    def transformations(self):
        return self._transformations
    
    def with_reader(self, value):
        self._reader = value
        return self

    def with_reader_and_parameters(self, reader_name:str, parameters:Tuple=()):
        self._reader = GenericReader(reader_name, parameters)
        return self

    def with_writer(self, value):
        self._writer = value
        return self

    def with_writer_and_parameters(self, writer_name:str, parameters:Tuple=()):
        self._writer = GenericWriter(writer_name, parameters)
        return self
    
    def clear_transformations(self):
        self._transformations = []
        return self
    
    def with_transformation(self, transformation):
        self._transformations.append(transformation)
        return self
    

class TransferSource:
    pass


class DatasetTransferSource(TransferSource):
    def __init__(self, dataset_id) -> None:
        super().__init__()
        self._dataset_id = dataset_id

    @property
    def dataset_id(self):
        return self._dataset_id


class LocalFileTransferSource(TransferSource):
    def __init__(self, file_path, original_file_name:str=None, max_parallelism:int=2, block_size:int=20*1024*1024) -> None:
        super().__init__()
        self._file_path = file_path
        self._original_file_name = original_file_name
        self._max_parallelism = max_parallelism
        self._block_size = block_size

    @property
    def file_path(self):
        return self._file_path

    @property
    def original_file_name(self):
        return self._original_file_name

    @property
    def max_parallelism(self):
        return self._max_parallelism

    @property
    def block_size(self):
        return self._block_size


class StreamTransferSource(TransferSource):
    def __init__(self, stream, original_file_name:str=None, max_parallelism:int=2, block_size:int=20*1024*1024) -> None:
        super().__init__()
        self._stream = stream
        self._original_file_name = original_file_name
        self._max_parallelism = max_parallelism
        self._block_size = block_size

    @property
    def stream(self):
        return self._stream

    @property
    def original_file_name(self):
        return self._original_file_name

    @property
    def max_parallelism(self):
        return self._max_parallelism

    @property
    def block_size(self):
        return self._block_size


class UrlTransferSource(TransferSource):
    def __init__(self, url:str, original_file_name:str=None) -> None:
        super().__init__()
        self._url = url
        self._original_file_name = original_file_name
    
    @property
    def url(self):
        return self._url

    @property
    def original_file_name(self):
        return self._original_file_name


class DownloadTransferOutput:
    def __init__(self, url) -> None:
        self._url = url

    @property
    def url(self):
        return self._url

    @classmethod
    def from_body(cls, body:dict):
        return cls(
            url = body["url"]
        )


class NewDatasetOutput:
    def __init__(self, project_id, dataset_id) -> None:
        self._project_id = project_id
        self._dataset_id = dataset_id

    @property
    def project_id(self):
        return self._project_id

    @property
    def dataset_id(self):
        return self._dataset_id

    @classmethod
    def from_body(cls, body:dict):
        return cls(
            project_id = body["projectId"],
            dataset_id = body["datasetId"]
        )


class EmptyTransferOutput:
    pass

    @classmethod
    def from_body(cls, body:dict):
        return cls()


class TransferProcessOutput:
    def __init__(self, transfer_id) -> None:
        self._transfer_id = transfer_id

    @property
    def transfer_id(self):
        return self._transfer_id
    
    @classmethod
    def from_body(cls, body:dict):
        return cls(body["transferId"])


class ConvertUploadInput:
    def __init__(
        self,
        original_file_name,
        upload_url,
        output_dataset_data: DatasetTransferInput,
        project_id,
        reader_parameters: Tuple,
        writer_parameters: Tuple,
        reader_name: str,
        writer_name: str,
        transformations: Tuple[TransformationBase]
    ) -> None:
        self._original_file_name = original_file_name
        self._upload_url = upload_url
        self._output_dataset_data = output_dataset_data
        self._project_id = project_id
        self._reader_parameters = list(reader_parameters)
        self._writer_parameters = list(writer_parameters)
        self._reader_name = reader_name
        self._writer_name = writer_name
        self._transformations = transformations

    @property
    def upload_url(self):
        return self._upload_url

    @upload_url.setter
    def upload_url(self, value):
        self._upload_url = value

    def body(self):
        body = {
            "originalFileName": self._original_file_name if self._original_file_name is not None else self._output_dataset_data.name,
            "uploadUrl": self._upload_url,
            "outputDatasetData": self._output_dataset_data.body(),
            "projectId": self._project_id,
            "readerName": self._reader_name,
            "writerName": self._writer_name
        }

        if self._transformations:
            body["transformations"] = [t.body() for t in self._transformations]
        
        if self._reader_parameters:
            body["readerParameters"] = [{"name": p[0], "value": p[1]} for p in self._reader_parameters]
        
        if self._writer_parameters:
            body["writerParameters"] = [{"name": p[0], "value": p[1]} for p in self._writer_parameters]

        return body

class ConvertFileUpdateInput:
    def __init__(
        self,
        original_file_name:str,
        upload_url:str,
        reader_name:str,
        writer_name:str,
        reader_parameters:Tuple=None,
        writer_parameters:Tuple=None,
        transformations:Tuple=None
    ) -> None:
        self._original_file_name = original_file_name
        self._upload_url = upload_url
        self._reader_name = reader_name
        self._writer_name = writer_name
        self._reader_parameters = reader_parameters
        self._writer_parameters = writer_parameters
        self._transformations = transformations
    
    @property
    def upload_url(self):
        return self._upload_url

    @upload_url.setter
    def upload_url(self, value):
        self._upload_url = value

    def body(self) -> dict:
        body = {
            "originalFileName": self._original_file_name,
            "uploadUrl": self._upload_url,
            "readerName": self._reader_name,
            "writerName": self._writer_name
        }
        
        if self._transformations:
            body["transformations"] = [t for t in self._transformations]

        if self._reader_parameters:
            body["readerParameters"] = [{"name": p[0], "value": p[1]} for p in self._reader_parameters]
        
        if self._writer_parameters:
            body["writerParameters"] = [{"name": p[0], "value": p[1]} for p in self._writer_parameters]
        
        return body

class ConvertInputBase:
    def __init__(
        self,
        reader_name:str,
        writer_name:str,
        reader_parameters:Tuple=None,
        writer_parameters:Tuple=None,
        transformations:Tuple[TransformationBase]=None
    ) -> None:
        self._reader_name = reader_name
        self._writer_name = writer_name
        self._reader_parameters = reader_parameters
        self._writer_parameters = writer_parameters
        self._transformations = transformations
    
    def body(self) -> dict:
        body = {
            "readerName": self._reader_name,
            "writerName": self._writer_name
        }
        
        if self._transformations:
            body["transformations"] = [t.body() for t in self._transformations]

        if self._reader_parameters:
            body["readerParameters"] = [{"name": p[0], "value": p[1]} for p in self._reader_parameters]
        
        if self._writer_parameters:
            body["writerParameters"] = [{"name": p[0], "value": p[1]} for p in self._writer_parameters]
        
        return body


class ConvertDatasetUpdateInput(ConvertInputBase):
    def __init__(self, reader_name: str, writer_name: str, reader_parameters: Tuple = None, writer_parameters: Tuple = None, transformations: Tuple = None) -> None:
        super().__init__(reader_name, writer_name, reader_parameters=reader_parameters, writer_parameters=writer_parameters, transformations=transformations)
    
    def body(self) -> dict:
        return super().body()


class ConvertDownloadInput(ConvertInputBase):
    def __init__(self, reader_name: str, writer_name: str, reader_parameters: Tuple = None, writer_parameters: Tuple = None, transformations: Tuple = None) -> None:
        super().__init__(reader_name, writer_name, reader_parameters=reader_parameters, writer_parameters=writer_parameters, transformations=transformations)
    
    def body(self) -> dict:
        return super().body()


class ConvertExistingInput(ConvertInputBase):
    def __init__(
        self,
        output_dataset_data:DatasetTransferInput,
        project_id,
        reader_name: str,
        writer_name: str,
        reader_parameters: Tuple = None,
        writer_parameters: Tuple = None,
        transformations: Tuple = None
    ) -> None:
        super().__init__(reader_name, writer_name, reader_parameters=reader_parameters, writer_parameters=writer_parameters, transformations=transformations)
        self._output_dataset_data = output_dataset_data
        self._project_id = project_id

    def body(self) -> dict:
        body = super().body()
        body["projectId"] = self._project_id
        body["outputDatasetData"] = self._output_dataset_data.body()
        return body


class TransferProcessAwaiter:
    def __init__(self, metadataclient2:MetadataClientV2):
        self._metadata2 = metadataclient2
    
    def wait_for_finish(self, transfer_id) -> TransferOutput:
        transfer_response = self._metadata2.GetTransferV2(transfer_id)
        transfer = TransferOutput.from_body(transfer_response.Body)

        while (transfer.status != TransferStatus.COMPLETED) and (transfer.status != TransferStatus.ERROR):
            time.sleep(2)
            transfer_response = self._metadata2.GetTransferV2(transfer_id)
            transfer = TransferOutput.from_body(transfer_response.Body)
        
        return transfer


class TransferProcessInvoker:
    def __init__(self, metadataclient2:MetadataClientV2, metadataclient3:MetadataClientV3, transfer_upload_helper:TransferUploadHelper):
        self._metadata2 = metadataclient2
        self._metadata3 = metadataclient3
        self._upload_helper = transfer_upload_helper
    
    def invoke(self, process):
        
        if isinstance(process, NewDatasetTransferProcess):

            transfer_source = process.transfer_source

            if isinstance(transfer_source, LocalFileTransferSource):
                input = self._process_to_convert_upload_input(process, "http://dummyUrl", transfer_source.original_file_name)
                
                url = self._upload_helper.stage_file(transfer_source.file_path)
                    
                input.upload_url = url

                response = self._metadata3.UploadConvertV3(input.body())
            
            elif isinstance(transfer_source, UrlTransferSource):
                input = self._process_to_convert_upload_input(process, transfer_source.url, transfer_source.original_file_name)

                response = self._metadata3.UploadConvertV3(input.body())

            elif isinstance(transfer_source, StreamTransferSource):
                input = self._process_to_convert_upload_input(process, "http://dummyUrl", transfer_source.original_file_name)

                url = self._upload_helper.stage_stream(transfer_source.stream, transfer_source.max_parallelism, transfer_source.block_size)

                input.upload_url = url
                
                response = self._metadata3.UploadConvertV3(input.body())

            elif isinstance(transfer_source, DatasetTransferSource):
                input = self._process_to_convert_existing_input(process)

                response = self._metadata3.ConvertDatasetV3(input.body(), transfer_source.dataset_id)

            else:
                raise NotImplementedError("Unsupported type of transfer source: " + transfer_source + " for process " + process)

            transfer = TransferSummaryOutput.from_body(response.Body)
            return TransferProcessOutput(transfer.id)
        
        elif isinstance(process, InPlaceDatasetTransferProcess):

            transfer_source = process.transfer_source
            update_source = process.update_source

            if isinstance(update_source, UrlTransferSource):
                input = self._process_to_convert_update_input(update_source.url, update_source.original_file_name, process)
                
                response = self._metadata3.UpdateFromFileV3(input.body(), transfer_source.dataset_id)
                transfer = TransferSummaryOutput.from_body(response.Body)

            elif isinstance(update_source, StreamTransferSource):
                input = self._process_to_convert_update_input("http://dummyUrl", update_source.original_file_name, process)

                url = self._upload_helper.stage_stream(update_source.stream, update_source.max_parallelism, update_source.block_size)

                input.upload_url = url
                
                response = self._metadata3.UpdateFromFileV3(input.body(), transfer_source.dataset_id)
                transfer = TransferSummaryOutput.from_body(response.Body)
            
            elif isinstance(update_source, LocalFileTransferSource):
                input = self._process_to_convert_update_input("http://dummyUrl", update_source.original_file_name, process)

                url = self._upload_helper.stage_file(update_source.file_path, update_source.max_parallelism, update_source.block_size)

                input.upload_url = url

                response = self._metadata3.UpdateFromFileV3(input.body(), transfer_source.dataset_id)
                transfer = TransferSummaryOutput.from_body(response.Body)

            elif isinstance(update_source, DatasetTransferSource):
                input = self._process_to_convert_dataset_upload_input(process)

                response = self._metadata3.UpdateFromDatasetV3(input.body(), transfer_source.dataset_id, update_source.dataset_id)
                transfer = TransferSummaryOutput.from_body(response.Body)

            else:
                raise NotImplementedError("Unsupported type of transfer source: " + transfer_source + " for process " + process)

            return TransferProcessOutput(transfer.id)
        
        elif isinstance(process, DownloadTransferProcess):
            input = self._process_to_convert_download_input(process)

            response = self._metadata3.DownloadConvertDatasetV3(input.body(), process.transfer_source.dataset_id)
            transfer = TransferSummaryOutput.from_body(response.Body)
            return TransferProcessOutput(transfer.id)

        else:
            
            raise NotImplementedError("Unsupported type of transfer process: " + str(process))

    def _process_to_convert_upload_input(self, process, url:str, original_name:str) -> ConvertUploadInput:
        reader_parameters = None
        reader_name = None
        writer_parameters = None
        writer_name = None
        transfer_pipeline = process.transfer_pipeline

        if hasattr(transfer_pipeline, "reader"):
            reader_parameters = getattr(transfer_pipeline.reader, "parameters", ())
            reader_name = transfer_pipeline.reader.name

        if hasattr(transfer_pipeline, "writer"):
            writer_parameters = getattr(transfer_pipeline.writer, "parameters", ())
            writer_name = transfer_pipeline.writer.name

        return ConvertUploadInput(
            original_file_name=original_name,
            upload_url=url,
            output_dataset_data=process.dataset_info,
            project_id=process.destination_project_id,
            reader_parameters=reader_parameters,
            reader_name=reader_name,
            writer_parameters=writer_parameters,
            writer_name=writer_name,
            transformations=transfer_pipeline.transformations
        )

    def _process_to_convert_update_input(self, url, original_name, process) -> ConvertFileUpdateInput:
        reader_parameters = None
        reader_name = None
        writer_parameters = None
        writer_name = None
        transfer_pipeline = process.transfer_pipeline

        if hasattr(transfer_pipeline, "reader"):
            reader_parameters = getattr(transfer_pipeline.reader, "parameters", ())
            reader_name = transfer_pipeline.reader.name

        if hasattr(transfer_pipeline, "writer"):
            writer_parameters = getattr(transfer_pipeline.writer, "parameters", ())
            writer_name = transfer_pipeline.writer.name

        return ConvertFileUpdateInput(
            original_file_name=original_name,
            upload_url=url,
            reader_name=reader_name,
            writer_name=writer_name,
            reader_parameters=reader_parameters,
            writer_parameters=writer_parameters,
            transformations=transfer_pipeline.transformations
        )

    def _process_to_convert_dataset_upload_input(self, process) -> ConvertDatasetUpdateInput:
        reader_parameters = None
        reader_name = None
        writer_parameters = None
        writer_name = None
        transfer_pipeline = process.transfer_pipeline

        if hasattr(transfer_pipeline, "reader"):
            reader_parameters = getattr(transfer_pipeline.reader, "parameters", ())
            #reader_name = transfer_pipeline.reader.name
            reader_name = getattr(transfer_pipeline.reader, "name", None)

        if hasattr(transfer_pipeline, "writer"):
            writer_parameters = getattr(transfer_pipeline.writer, "parameters", ())
            #writer_name = transfer_pipeline.writer.name
            writer_name = getattr(transfer_pipeline.writer, "name", None)
            # HERE

        return ConvertDatasetUpdateInput(
            reader_name=reader_name,
            writer_name=writer_name,
            reader_parameters=reader_parameters,
            writer_parameters=writer_parameters,
            transformations=transfer_pipeline.transformations
        )

    def _process_to_convert_download_input(self, process) -> ConvertDownloadInput:
        reader_parameters = None
        reader_name = None
        writer_parameters = None
        writer_name = None
        transfer_pipeline = process.transfer_pipeline

        if hasattr(transfer_pipeline, "reader"):
            reader_parameters = getattr(transfer_pipeline.reader, "parameters", ())
            reader_name = transfer_pipeline.reader.name

        if hasattr(transfer_pipeline, "writer"):
            writer_parameters = getattr(transfer_pipeline.writer, "parameters", ())
            writer_name = transfer_pipeline.writer.name

        return ConvertDownloadInput(
            reader_name=reader_name,
            writer_name=writer_name,
            reader_parameters=reader_parameters,
            writer_parameters=writer_parameters,
            transformations=transfer_pipeline.transformations
        )

    def _process_to_convert_existing_input(self, process) -> ConvertExistingInput:
        reader_parameters = None
        reader_name = None
        writer_parameters = None
        writer_name = None
        transfer_pipeline = process.transfer_pipeline

        if hasattr(transfer_pipeline, "reader"):
            reader_parameters = getattr(transfer_pipeline.reader, "parameters", ())
            reader_name = transfer_pipeline.reader.name

        if hasattr(transfer_pipeline, "writer"):
            writer_parameters = getattr(transfer_pipeline.writer, "parameters", ())
            writer_name = transfer_pipeline.writer.name

        dataset_info = process.dataset_info

        dataset_input = DatasetTransferInput(
            name=dataset_info.name,
            description=dataset_info.description,
            metadata=dataset_info.metadata,
            properties=dataset_info.properties
        )

        return ConvertExistingInput(
            output_dataset_data = dataset_input,
            project_id=process.destination_project_id,
            reader_name = reader_name,
            writer_name = writer_name,
            reader_parameters = reader_parameters,
            writer_parameters = writer_parameters,
            transformations=transfer_pipeline.transformations
        )


class TransferProcess:
    def __init__(
        self,
        transfer_source:TransferSource,
        transfer_pipeline:TransferPipeline,
        awaiter:TransferProcessAwaiter
    ) -> None:
        self._transfer_source = transfer_source
        self._transfer_pipeline = transfer_pipeline
        self._awaiter = awaiter

    @property
    def transfer_pipeline(self) -> TransferPipeline:
        return self._transfer_pipeline

    @property
    def transfer_source(self):
        return self._transfer_source

    def execute_and_wait(self):
        transfer_process_output = self.execute()
        final_output = self._awaiter.wait_for_finish(transfer_process_output.transfer_id)
        
        if final_output.status == TransferStatus.ERROR:
            raise MikeCloudException(f"Transfer {final_output.id} finished with error: {final_output.error_message}")

        return self.create_transfer_result(final_output)

    def execute(self) -> TransferProcessOutput:
        self.validate_before_execution()
        return self.execute_self()
    
    @abstractmethod
    def validate_before_execution(self):
        pass

    @abstractmethod
    def create_transfer_result(self, transfer_output: TransferOutput):
        pass

    @abstractmethod
    def execute_self(self) -> TransferProcessOutput:
        pass

    def with_reader_and_parameters(self, reader_name:str, parameters:Tuple=()):
        self._transfer_pipeline = self._transfer_pipeline.with_reader_and_parameters(reader_name, parameters)
        return self

    def with_reader(self, reader):
        self._transfer_pipeline = self._transfer_pipeline.with_reader(reader)
        return self

    def with_writer_and_parameters(self, writer_name:str, parameters:Tuple=()):
        self._transfer_pipeline = self._transfer_pipeline.with_writer_and_parameters(writer_name, parameters)
        return self

    def with_writer(self, writer):
        self._transfer_pipeline = self._transfer_pipeline.with_writer(writer)
        return self
    
    def with_transformation(self, transformation):
        self._transfer_pipeline = self._transfer_pipeline.with_transformation(transformation)
        return self
    
    def _with_specific_reader(self, reader, configure_reader=None):
        if configure_reader and callable(configure_reader):
            configure_reader(reader)
        self.with_reader(reader)
        return self

    def _with_specific_writer(self, writer, configure_writer=None):
        if configure_writer and callable(configure_writer):
            configure_writer(writer)
        self.with_writer(writer)
        return self

    def with_dfs2_reader(self, configure_reader=None):
        """
        :param configure_reader: Optional callable to configure reader properties, e.g. to add parameters
        """
        return self._with_specific_reader(Dfs2Reader(), configure_reader)

    def with_dfsu_reader(self, configure_reader=None):
        """
        :param configure_reader: Optional callable to configure reader properties, e.g. to add parameters
        """
        return self._with_specific_reader(DfsuReader(), configure_reader)

    def with_multidimensional_writer(self, configure_writer=None):
        """
        :param configure_writer: Optional callable to configure writer properties, e.g. to add parameters
        """
        return self._with_specific_writer(MDWriter(), configure_writer)

    def with_dfs2_writer(self, configure_writer=None):
        """
        :param configure_writer: Optional callable to configure writer properties, e.g. to add parameters
        """
        return self._with_specific_writer(Dfs2Writer(), configure_writer)

    def with_timeseries_writer(self, configure_writer=None):
        """
        :param configure_writer: Optional callable to configure writer properties, e.g. to add parameters
        """
        return self._with_specific_writer(TSWriter(), configure_writer)

    def with_coordinate_system_transformation(self, output_srid:int, override_input_srid:int=None):
        """
        Include coordinate system transformation in the transfer pipeline.

        :param output_srid: Desired Spatial Reference ID of output
        :param override_input_srid: Optional Spatial Reference ID of input if different than automatically detected SRID should be used
        """
        crs_transformation = CrsTransformation(output_srid, override_input_srid)
        self._transfer_pipeline.with_transformation(crs_transformation)
        return self
    
    def with_item_filtering_transformation(self, item_indices:Tuple[int]):
        """
        Include item filtering transformation in the transfer pipeline.

        :param item_indices: Indices of the items to be included in the output
        """
        item_filter_transformation = ItemFilterTransformation(item_indices)
        self._transfer_pipeline.with_transformation(item_filter_transformation)
        return self

    def with_time_filtering_transformation(self, from_:datetime.datetime=None, to:datetime.datetime=None):
        """
        Include temporal filtering transformation in the transfer pipeline.

        :param from_: Include time slices after this time
        :param to: Include time slices before this time
        """
        if not (from_ or to):
            raise MikeCloudException("At least one of from_ or to parameters must be specified")

        if (from_ and to) and  (to < from_):
            raise MikeCloudException("Parameter from_ must be lower than parameter to")

        temporal_value_filter = TemporalValueFilter(from_, to)
        temporal_filter_transformation = TemporalFilterTransformation(temporal_value_filter)
        self._transfer_pipeline.with_transformation(temporal_filter_transformation)
        return self

    def with_spatial_filtering_transformation(self, geometry:str, srid:int=None):
        """
        Include spatial filtering transformation in the transfer pipeline.

        :param geometry: WKT polygon to select data within
        :param srid: Spatial Reference ID of the geometry if diffrent from data Spatial Reference ID
        """
        spatial_filter = SpatialFilter(geometry, srid)
        spatial_transformation = SpatialFilterTransformation(spatial_filter)
        self._transfer_pipeline.with_transformation(spatial_transformation)
        return self
    

class InPlaceDatasetTransferProcess(TransferProcess):
    def __init__(
        self, 
        transfer_source: TransferSource, 
        update_source: TransferSource,
        transfer_pipeline: TransferPipeline, 
        awaiter: TransferProcessAwaiter,
        invoker: TransferProcessInvoker
    ) -> None:
        super().__init__(transfer_source, transfer_pipeline, awaiter)
        self._update_source = update_source
        self._invoker = invoker
    
    @property
    def update_source(self) -> TransferSource:
        return self._update_source
    
    def create_transfer_result(self, transfer_output:TransferOutput) -> EmptyTransferOutput:
        return EmptyTransferOutput()
    
    def execute_self(self) -> TransferProcessOutput:
        return self._invoker.invoke(self)


class DownloadTransferProcess(TransferProcess):
    def __init__(
        self, 
        transfer_source: TransferSource, 
        transfer_pipeline: TransferPipeline,
        awaiter: TransferProcessAwaiter,
        invoker: TransferProcessInvoker,
    ) -> None:
        super().__init__(transfer_source, transfer_pipeline, awaiter)
        self._invoker = invoker

    def create_transfer_result(self, transfer_output:TransferOutput) -> DownloadTransferOutput:
        return DownloadTransferOutput(transfer_output.download_path)

    def execute_self(self) -> TransferProcessOutput:
        return self._invoker.invoke(self)


class NewDatasetTransferProcess(TransferProcess):
    def __init__(
        self,
        transfer_source: TransferSource,
        transfer_pipeline:TransferPipeline,
        destination_project_id,
        dataset_info:DatasetTransferInput,
        awaiter: TransferProcessAwaiter,
        invoker: TransferProcessInvoker
    ) -> None:
        super().__init__(transfer_source, transfer_pipeline, awaiter)
        self._destination_project_id = destination_project_id
        self._dataset_info = dataset_info
        self._invoker = invoker
    
    @property
    def dataset_info(self) -> DatasetTransferInput:
        return self._dataset_info

    @dataset_info.setter
    def dataset_info(self, value):
        self._dataset_info = value

    @property
    def transfer_source(self):
        return self._transfer_source

    @property
    def destination_project_id(self):
        return self._destination_project_id

    def with_dataset_description(self, description:str):
        self._dataset_info.description = description
        return self
    
    def with_dataset_name(self, name:str):
        self._dataset_info.name = name
        return self

    def with_dataset_metadata(self, name:str, description:str, metadata:dict=None, properties:dict=None):
        self.dataset_info = DatasetTransferInput(name, description, metadata, properties)
        return self

    def create_transfer_result(self, transfer_output:TransferOutput) -> NewDatasetOutput:
        if transfer_output.import_results and len(transfer_output.import_results) == 1:
            import_result = transfer_output.import_results[0]
            return NewDatasetOutput(import_result.project_id, import_result.dataset_id)
        
        raise MikeCloudException("Expected at least one import result of a transfer")
    
    def execute_self(self) -> TransferProcessOutput:
        return self._invoker.invoke(self)


class TransferClient():
    
    def __init__(self, inspectFnc=MetadataGenClientV1.DefaultInspectFnc, **kwargs):
        self._metadata2 = kwargs.get("MetadataClientV2", MetadataClientV2(inspectFnc, **kwargs))
        self._metadata3 = kwargs.get("MetadataClientV3", MetadataClientV3(inspectFnc, **kwargs))
        self._raw2 = kwargs.get("RawClientV2", RawClientV2(inspectFnc, **kwargs))
        self._transfer_upload_helper = TransferUploadHelper(self._metadata2)
        self._transfer_process_invoker = TransferProcessInvoker(self._metadata2, self._metadata3, self._transfer_upload_helper)
        self._transfer_process_awaiter = TransferProcessAwaiter(self._metadata2)

    def upload_file(self, local_file_path:str, project_id:str, name:str=None, description:str=None, metadata:dict=None, verbose:bool=False, timeout=60, reporter=lambda message: print(message)) -> str:
        """
        Upload a file from local path to a project in MIKE Cloud Platform.
        :param local_file_path: Path to the file to upload.
        :param project_id: ID of the project to upload to.
        :param name: Desired name of the resulting dataset, default is None for name of the file. Platform naming restrictions apply.
        :param description: Description of the dataset, default is None for no description
        :param metadata: Dictionary of any relevant metadata.
        :param verbose: True if this call should report messages, otherwise False.
        :param reporter: Function for reporting messages, takes one parameter - the message.
        :return: Dataset id of the uploaded file.
        :rtype: str
        """

        if(verbose):
            reporter("Preparing upload {}...".format(local_file_path))
        blob_url = self._metadata2.GetUploadUrlV2().Body["data"]
        blob = BlobClient.from_blob_url(blob_url)
        
        if(verbose):
            reporter("Uploading to {}...".format(blob_url))
        with open(local_file_path, "rb") as data:
            blob.upload_blob(data)
        
        if(verbose):
            reporter("Uploaded as {}".format(blob.url))
            reporter("Importing...")
        
        file_name = os.path.basename(local_file_path)
        
        output_dataset_data = { "name": file_name }
        if description:
            output_dataset_data["description"] = description
        if metadata:
            output_dataset_data["metadata"] = metadata
        
        upload_convert_input = {
            "originalFileName": file_name,
            "uploadUrl": blob_url,
            "outputDatasetData": output_dataset_data,
            "readerName": "FileReader",
            "writerName": "FileWriter",
            "projectId": project_id
        }
        
        conversion = self._metadata3.UploadConvertV3(upload_convert_input)
        transfer_id = conversion.Body["id"]

        start_time = datetime.datetime.now()
        
        waiting_times = (2, 3, 5, 8, 13, 21)
        i = 0
        while(True):
            t = datetime.datetime.now()
            
            if (start_time - t).total_seconds() >= timeout:
                raise MikeCloudException(f"Upload timed out for {local_file_path}")
            
            transfer = self._metadata2.GetTransferV2(transfer_id)
            status = transfer.Body["status"].lower()
            
            if(verbose):
                reporter(f"transfer {transfer_id} {status}")
            

            pause = waiting_times[i] if i < len(waiting_times) else waiting_times[-1]
            time.sleep(pause)
            i += 1

            if(transfer.Body["status"].lower() == "completed"):
                
                dataset_id = transfer.Body["importResults"][0]["datasetId"]
                
                if(verbose):
                    reporter(f"Imported dataset id is {dataset_id}")
                
                return dataset_id
    
    def download_file(self, project_id:str, dataset_id:str, local_file_destination:str, mode:str="w", verbose:bool=False, reporter=lambda message: print(message)) -> str:
        """
        Download dataset from a project to a local file.

        Only dataset type 'file' can be downloaded this way, no conversion is available.

        :param project_id: ID of the project to download data from
        :param dataset_id: ID of the dataset to download
        :param local_file_destination: file path to download to
        :param mode: File mode - "x" for fail if local file already exists, default is "w" for owewrite local file if already exists
        :return: path to the downloaded file
        :rtype: str
        """

        mode = mode.lower()
        if mode not in ('w', 'x', ):
            raise MikeCloudException(f"Invalid parameter mode, must be 'w' or 'x', not {mode}.")
        
        if(verbose):
            reporter("Preparing download...")
        
        response = self._raw2.GetFileSasUrlV2(project_id, dataset_id)
        blob_url = response.Body["data"]
        blob = BlobClient.from_blob_url(blob_url)
        
        if(verbose):
            reporter("Downloading...")
        
        mode += "b"
        with open(local_file_destination, mode) as f:
            download_stream = blob.download_blob()
            f.write(download_stream.readall())
        
        return local_file_destination

    def get_readers(self, filter:ConverterFilter=ConverterFilter.ALL) -> Generator[ConverterOutput, None, None]:
        response = self._metadata2.GetReadersListV2(filter.name.title())
        for d in response.Body["data"]:
            yield ConverterOutput.from_body(d)

    def get_writers(self, filter:ConverterFilter=ConverterFilter.ALL):
        response = self._metadata2.GetWritersListV2(filter.name.title())
        for d in response.Body["data"]:
            yield ConverterOutput.from_body(d)

    def get_transfer(self, transfer_id:uuid) -> TransferOutput:
        response = self._metadata2.GetTransferV2(transfer_id)
        return TransferOutput.from_body(response.Body)

    def stage_file(self, file_path:str, max_parallelism=2, block_size=20*1024*1024) -> str:
        """
        Prepare a file for import into the platform by uploading it into a staging area.

        :param file_path: Input file
        :param max_parallelism: maximum concurrency to use when files are over 64 MB
        :param block_size: max_block_size - The maximum chunk size for uploading a block blob in chunks.
        :return: Url with the staged file (the url can be used as an input for transfer operations)
        :rtype: str
        """
        return self._transfer_upload_helper.stage_file(file_path, max_parallelism, block_size) 
    
    def stage_stream(self, stream, max_parallelism:int=2, block_size:int=20*1024*1024) -> str:
        """
        Prepare a stream for import into the platform by uploading it into a staging area.

        :param stream: Input stream
        :param max_parallelism: maximum concurrency to use when files are over 64 MB
        :param block_size: max_block_size - The maximum chunk size for uploading a block blob in chunks.
        :return: Url with the staged stream (the url can be used as an input for transfer operations)
        :rtype: str
        """
        return self._transfer_upload_helper.stage_stream(stream, max_parallelism, block_size)

    
    def create_file_import(self, destination_project_id, file_path:str, max_parallelism:int=2, block_size:int=20*1024*1024) -> NewDatasetTransferProcess:
        """
        Creates a file import task based on a local file. The import will create a new dataset when executed.
        :param destination_project_id: Project id where the new dataset will end up.
        :param file_path: Path to the local file
        :param max_parallelism: Optional settings controlling the physical upload
        :param block_size: Optional settings controlling the physical upload
        :returns: Object representing the import task. Has to be executed in order for it to start.
        :rtype: NewDatasetTransferProcess
        """
        source = LocalFileTransferSource(file_path, None, max_parallelism=max_parallelism, block_size=block_size)
        pipeline = TransferPipeline(FileReader(), FileWriter())
        dataset_transfer_input = DatasetTransferInput(os.path.basename(file_path), None, None, None)

        return NewDatasetTransferProcess(
            transfer_source=source,
            transfer_pipeline=pipeline,
            destination_project_id=destination_project_id,
            dataset_info=dataset_transfer_input,
            invoker=self._transfer_process_invoker,
            awaiter=self._transfer_process_awaiter
        )


    def create_stream_import(self, destination_project_id, stream, dataset_name:str, original_file_name:str=None, max_parallelism:int=2, block_size:int=20*1024*1024) -> NewDatasetTransferProcess:
        """
        Creates a file import task based on a stream. The import will create a new dataset when executed.

        :param destination_project_id: Project id where the new dataset will end up.
        :param stream: Input stream
        :param dataset_name: Name of the output dataset
        :param original_file_name: Name representing the stream as if it was a file (the name could be important because of the file extension)
        :param max_parallelism: Optional settings controlling the physical upload
        :param block_size: Optional settings controlling the physical upload
        :returns: Object representing the import task. Has to be executed in order for it to start.
        :rtype: NewDatasetTransferProcess
        """
        source = StreamTransferSource(stream, original_file_name, max_parallelism, block_size)
        pipeline = TransferPipeline(FileReader(), FileWriter())
        dataset_transfer_input = DatasetTransferInput(dataset_name)
        
        return NewDatasetTransferProcess(
            transfer_source=source,
            transfer_pipeline=pipeline,
            destination_project_id=destination_project_id,
            dataset_info=dataset_transfer_input,
            invoker=self._transfer_process_invoker,
            awaiter=self._transfer_process_awaiter
        )


    def create_url_import(self, destination_project_id, remote_url:str, dataset_name:str, original_file_name:str=None) -> NewDatasetTransferProcess:
        """
        Creates a file import task based on a url. The import will create a new dataset when executed.

        :param destination_project_id: Project id where the new dataset will end up.
        :param remote_url: Url representing the file to be imported
        :param dataset_name: Name of the output dataset
        :param original_file_name: Name representing the stream as if it was a file (the name could be important because of the file extension)
        :returns: Object representing the import task. Has to be executed in order for it to start.
        :rtype: NewDatasetTransferProcess
        """
        source = UrlTransferSource(remote_url, original_file_name)
        pipeline = TransferPipeline(FileReader(), FileWriter())
        dataset_transfer_input = DatasetTransferInput(dataset_name)
        
        return NewDatasetTransferProcess(
            transfer_source=source,
            transfer_pipeline=pipeline,
            destination_project_id=destination_project_id,
            dataset_info=dataset_transfer_input,
            invoker=self._transfer_process_invoker,
            awaiter=self._transfer_process_awaiter
        )


    def create_dataset_copy(self, source_project_id, source_dataset_id, destination_project_id, dataset_name) -> NewDatasetTransferProcess:
        """
        Creates a dataset copy/conversion task. When executed, the task will create a new dataset based on an existing dataset.

        :param source_project_id: Parent project id of the source dataset
        :param source_dataset_id: Id of the source dataset
        :param destination_project_id: Project id for the new dataset
        :param dataset_name: Name of the new dataset        
        :returns: Object representing the copy/convert task. Has to be executed in order for it to start.
        :rtype: NewDatasetTransferProcess
        """
        source = DatasetTransferSource(source_dataset_id)
        pipeline = TransferPipeline(None, None)
        dataset_transfer_input = DatasetTransferInput(dataset_name)
        
        return NewDatasetTransferProcess(
            transfer_source=source,
            transfer_pipeline=pipeline,
            destination_project_id=destination_project_id,
            dataset_info=dataset_transfer_input,
            invoker=self._transfer_process_invoker,
            awaiter=self._transfer_process_awaiter
        )

    def create_dataset_download(self, source_project_id, source_dataset_id) -> DownloadTransferProcess:
        """
        Creates a dataset download/convert task. When executed, the task will create a url where the output of the conversion can be downloaded.

        :param source_project_id: Parent project id of the source dataset
        :param source_dataset_id: Id of the source dataset
        :returns: Object representing the download/convert task. Has to be executed in order for it to start.
        :rtype: DownloadTransferProcess
        """
        source = DatasetTransferSource(source_dataset_id)
        pipeline = TransferPipeline(None, FileWriter())
        return DownloadTransferProcess(source, pipeline, self._transfer_process_awaiter, self._transfer_process_invoker)

    def create_dataset_update_from_file(self, project_id, dataset_id, file_path:str, original_file_name:str, max_parallelism:int=2, block_size:int=20*1024*1024) -> InPlaceDatasetTransferProcess:
        """
        Creates a dataset download/convert task. When executed, the task will create a url where the output of the conversion can be downloaded.

        :param project_id: Parent project id of the target dataset
        :param dataset_id: Id of the target dataset
        :param file_path: Path to the local file
        :param max_parallelism: Optional settings controlling the physical upload
        :param block_size: Optional settings controlling the physical upload
        :returns: Object representing the dataset update task. Has to be executed in order for it to start.
        :rtype: InPlaceDatasetTransferProcess 
        """
        transfer_source = DatasetTransferSource(dataset_id)
        pipeline = TransferPipeline(None, None)
        update_source = LocalFileTransferSource(file_path, original_file_name, max_parallelism, block_size)
        return InPlaceDatasetTransferProcess(
            transfer_source,
            update_source,
            pipeline,
            self._transfer_process_awaiter,
            self._transfer_process_invoker
        )
    
    def create_dataset_update_from_stream(self, project_id, dataset_id, stream, original_file_name:str, max_parallelism:int=2, block_size:int=20*1024*1024) -> InPlaceDatasetTransferProcess:
        """
        Creates a dataset update task based on a stream. When executed, the task will update an existing dataset with the contents of the source file.

        :param project_id: Parent project id of the target dataset
        :param dataset_id: Id of the target dataset
        :param stream: Stream containing the data
        :param max_parallelism: Optional settings controlling the physical upload
        :param block_size: Optional settings controlling the physical upload
        :returns: Object representing the dataset update task. Has to be executed in order for it to start.
        :rtype: InPlaceDatasetTransferProcess 
        """
        transfer_source = DatasetTransferSource(dataset_id)
        pipeline = TransferPipeline(None, None)
        update_source = StreamTransferSource(stream, original_file_name, max_parallelism, block_size)
        return InPlaceDatasetTransferProcess(
            transfer_source,
            update_source,
            pipeline,
            self._transfer_process_awaiter,
            self._transfer_process_invoker
        )

    def create_dataset_update_from_url(self, project_id, dataset_id, remote_url:str, original_file_name:str) -> InPlaceDatasetTransferProcess:
        """
        Creates a dataset update task based on a remote url. When executed, the task will update an existing dataset with the contents of the source file.

        :param project_id: Parent project id of the target dataset
        :param dataset_id: Id of the target dataset
        :param remote_url: Url containing the data
        :param max_parallelism: Optional settings controlling the physical upload
        :param block_size: Optional settings controlling the physical upload
        :returns: Object representing the dataset update task. Has to be executed in order for it to start.
        :rtype: InPlaceDatasetTransferProcess 
        """
        transfer_source = DatasetTransferSource(dataset_id)
        pipeline = TransferPipeline(None, None)
        update_source = UrlTransferSource(remote_url, original_file_name)
        return InPlaceDatasetTransferProcess(
            transfer_source,
            update_source,
            pipeline,
            self._transfer_process_awaiter,
            self._transfer_process_invoker
        )

    def create_dataset_update_from_dataset(self, source_dataset_id, target_dataset_id) -> InPlaceDatasetTransferProcess:
        """
        Creates a dataset update task based on another dataset. When executed, the task will update an existing dataset with the contents of another dataset.
        
        :param source_dataset_id: Id of the source dataset providing updating data
        :param target_dataset_id: Id of the target dataset to be updated
        :return: Object representing the dataset update task. Has to be executed in order for it to start.
        :rtype: InPlaceDatasetTransferProcess
        """
        transfer_source = DatasetTransferSource(target_dataset_id)
        pipeline = TransferPipeline(None, None)
        update_source = DatasetTransferSource(source_dataset_id)
        return InPlaceDatasetTransferProcess(
            transfer_source,
            update_source,
            pipeline,
            self._transfer_process_awaiter,
            self._transfer_process_invoker
        )

    def list_project_transfers(self, project_id, from_:datetime.datetime=None, to:datetime.datetime=None, statuses:Tuple[TransferStatus]=()) -> Generator[TransferSummaryOutput, None, None]:
        """
        Yield transfers for specified project
        :param project_id: Id of the project
        :param from_: filter only transfers after this time
        :param to: filter only transfers before this time
        :param statuses: filter only specified transfer statuses
        :return: transfer summaries
        :rtype: Generator[TransferSummaryOutput, None, None]        
        """
        limit = 1000
        offset = 0
        first_query = True
        while True:
            if not first_query:
                if not response.Body["data"]:
                    break
            response = self._metadata2.GetProjectTransferListV2(project_id, from_=from_, to=to, status=statuses, offset=offset, limit=limit, datasetid=None)
            for i in response.Body["data"]:
                yield TransferSummaryOutput.from_body(i)
            offset = offset + limit + 1
            first_query = False
    
    def list_dataset_transfers(self, project_id, dataset_id, from_:datetime.datetime=None, to:datetime.datetime=None, statuses:Tuple[TransferStatus]=()) -> Generator[TransferSummaryOutput, None, None]:
        """
        Yield transfers for specified dataset
        :param project_id: Id of the project that holds the dataset
        :param dataset_id: Id of the dataset
        :param from_: filter only transfers after this time
        :param to: filter only transfers before this time
        :param statuses: filter only specified transfer statuses
        :return: transfer summaries
        :rtype: Generator[TransferSummaryOutput, None, None]
        """
        limit = 1000
        offset = 0
        first_query = True
        while True:
            if not first_query:
                if not response.Body["data"]:
                    break
            response = self._metadata2.GetProjectTransferListV2(projectid=project_id, from_=from_, to=to, status=statuses, offset=offset, limit=limit, datasetid=dataset_id)
            for i in response.Body["data"]:
                yield TransferSummaryOutput.from_body(i)
            offset = offset + limit + 1
            first_query = False

    def is_dataset_locked(self, project_id, dataset_id) -> bool:
        """
        Determines whether a dataset is locked by a running transfer
        
        :param project_id: Project id
        :param dataset_id: Dataset id
        :return: True if dataset is locked by a running transfer, false otherwise
        :rtype: bool
        """
        transfers = self.list_dataset_transfers(project_id, dataset_id, statuses = (TransferStatus.NONE, TransferStatus.PENDING, TransferStatus.INPROGRESS))
        return bool(next(transfers, None))

    def upload_staged_files(self, project_id, input:StagedFilesUploadInput) -> StagedFilesUploadOutput:
        """
        Uploads files prepared in stage blobs
        
        :param project_id: Project id
        :param input: Info about uploaded files
        :return: Info about succeeded and failed uploads
        :rtype: StagedFilesUploadOutput
        """
        response = self._metadata2.UploadStagedFilesV2(input.body(), project_id)
        return StagedFilesUploadOutput.from_body(response.Body)


if __name__ == '__main__':
    print(__file__)
    print(dir())


# Copyright (c) 2021 DHI A/S - DHI Water Environment Health 
# All rights reserved.
# 
# This code is licensed under the MIT License.
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files(the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and / or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

class MikeCloudException(Exception):
    """Exception raised by MIKE Cloud Python SDK"""
    def __init__(self, *args: object) -> None:
        super().__init__(*args)

class MikeCloudRestApiException(MikeCloudException):
    """Exception raised when MIKE Cloud REST API call fails"""
    def __init__(self, message, detail, status_code, *args: object) -> None:
        super().__init__(message, *args)
        self._message = message
        self._status_code = status_code
        self._detail = detail
    
    @property
    def message(self):
        return self._message

    @property
    def status_code(self):
        return self._status_code
    
    @property
    def detail(self):
        return self._detail
    
    def __repr__(self) -> str:
        return f"{type(self).__name__}, HTTP ERROR {self.status_code}, {self.message}, {self.detail}"
    
    def __str__(self) -> str:
        return self.__repr__() + " " + super().__str__()


#!/usr/bin/env python
from dhi.platform.args import ClientArgs
from dhi.platform.config import ClientConfig
from dhi.platform.generated.metadatagen import MetadataGenClientV2
from dhi.platform.fmt import Format

def main():
    args = ClientArgs.ParseForDatasetList(description="Get details about deleted dataset", defaultformat="yaml")
    ClientConfig.UpdateProjectFromConfiguration(args)
    clientv2 = MetadataGenClientV2(**vars(args))

    responses = (clientv2.GetRecyclableDatasetV2(id) for id in args.datasetids)

    tablefmt = "{!s:32}\t{!s:32}\t{!s:16}\t{!s:11}\t{!s:13}\t{!s:11}\t{!s:26}\t{}"
    tablefields = ["id", "projectId", "name", "datasetType", "datasetFormat", "storageSize", "createdAt", "updatedAt"]
    Format.FormatResponses(responses, lambda r: r.Body, args.format, tablefmt, tablefields)

if __name__ == '__main__':
    main()



# Copyright (c) 2021 DHI A/S - DHI Water Environment Health 
# All rights reserved.
# 
# This code is licensed under the MIT License.
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files(the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and / or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
from .._version import VERSION

DHI_PYTHONSDK_VERSION = VERSION
DATETIMEFORMAT        = '%Y-%m-%dT%H:%M:%S.%f'
DATETIMEFORMAT_SECONDS= '%Y-%m-%dT%H:%M:%S'
DATETIMEFORMAT_SECONDS_URL= '%Y-%m-%dT%H%M%S'


#!/usr/bin/env python
from dhi.platform.args import ClientArgs
from dhi.platform.config import ClientConfig
from dhi.platform.metadatahelper import MetadataClientV3Contracts
from dhi.platform.generated.metadatagen import MetadataGenClientV3
from dhi.platform.fmt import Format

def __initParser(parser):
    parser.add_argument("-f", "--inputfile", help="Input file path, '-' to read stdin")
    #format: string
    #projectId: 3fa85f64-5717-4562-b3fc-2c963f66afa6
    #appendDatasetId: 3fa85f64-5717-4562-b3fc-2c963f66afa6
    #uploadUrl: string
    #fileName: string
    #srid: 0
    #arguments: {}
    #destinations:
    #  - Dedicated
    #datasetImportData:
    #  name: string
    #  description: string
    #  metadata: {}
    #  properties: {}
    parser.add_argument("--inputformat", help="Input format")
    parser.add_argument("--appenddatasetid", help="Append dataset id")
    parser.add_argument("--uploadurl", help="Upload URL")
    parser.add_argument("--filename", help="File name")
    parser.add_argument("--srid", help="SRID")
    parser.add_argument("--arguments", help="Arguments JSON, {}")
    parser.add_argument("--destinations", default="[\"Dedicated\"]", help="Destinations array JSON, []")
    parser.add_argument("--name", help="Name")
    parser.add_argument("--description", help="Description")
    parser.add_argument("--metadata", help="Dataset metadata JSON, {}")
    parser.add_argument("--properties", help="Dataset properties JSON, {}")

def main():
    args = ClientArgs.ParseForProject(description="Upload file", init=__initParser, defaultformat="yaml")
    ClientConfig.UpdatePlatformFromConfiguration(args)
    clientv3 = MetadataGenClientV3(**vars(args))

    input = MetadataClientV3Contracts.PrepareUploadInput(
        ClientArgs.LoadJson(args.inputfile),
        args.projectid,
        args.inputformat,
        args.appenddatasetid,
        args.uploadurl,
        args.filename,
        args.srid,
        ClientArgs.LoadJsonStr(args.arguments),
        ClientArgs.LoadJsonStr(args.destinations),
        args.name if args.name else args.filename,
        args.description,
        ClientArgs.LoadJsonStr(args.metadata),
        ClientArgs.LoadJsonStr(args.properties))
    response = clientv3.UploadV3(input)

    Format.FormatResponse(response, format=args.format)

if __name__ == '__main__':
    main()



#!/usr/bin/env python
from dhi.platform.args import ClientArgs
from dhi.platform.config import ClientConfig
from dhi.platform.generated.metadatagen import MetadataGenClientV3
from dhi.platform.fmt import Format

def main():
    args = ClientArgs.ParsePlatform(description="List services", defaultformat="yaml")
    ClientConfig.UpdatePlatformFromConfiguration(args)
    clientv3 = MetadataGenClientV3(**vars(args))

    response = clientv3.GetServiceIds()

    tablefmt = "{}"
    Format.FormatResponse(response, lambda r: r.Body.get("data"), args.format, tablefmt, None)

if __name__ == '__main__':
    main()



#!/usr/bin/env python
import asyncio, platform
from dhi.platform.args import ClientArgs
from dhi.platform.config import ClientConfig
from dhi.platform.eventinghelper import ExecutionSession
from dhi.platform.generated.enginegen import EngineGenClientV2
from dhi.platform.generated.metadatagen import MetadataGenClientV2, MetadataGenClientV3
from dhi.platform.fmt import Format

def __initParser(parser):
    parser.add_argument("-f", "--inputfile", default=ClientArgs.GetDefault("DHIINPUT"), help="Input file path")
    parser.add_argument("-w", "--wait", default=False, help="Wait to finish", action="store_true")
    parser.add_argument("-l", "--showlog", default=False, help="Show log", action="store_true")

async def main():
    args = ClientArgs.ParseForProject(description="Run engine", init=__initParser, defaultformat="yaml")
    ClientConfig.UpdateProjectFromConfiguration(args)
    clientv2 = EngineGenClientV2(**vars(args))

    metadataclientv2 = MetadataGenClientV2(**vars(args))
    metadataclientv3 = MetadataGenClientV3(**vars(args))
    executionsession = ExecutionSession(args.showlog, metadataclientv2, metadataclientv3, args.verbose)
    async with executionsession.subscribe(args.projectid, ["/dhi/platform/engineexecution"], 200, None):
        try:
            await executionsession.wait_pubsubconnected()

            input = ClientArgs.LoadJson(args.inputfile)
            if args.showlog:
                modelitems = input.get("models")
                if modelitems:
                    for i in modelitems:
                        if i.get("engine"):
                            tmp = i.get("reportLogUpdatesLines")
                            if not tmp or tmp == 0:
                                i["reportLogUpdatesLines"] = 300
            response = clientv2.RunExecutionWithPlatformData(args.projectid, input)

            executionsession.set_resourceid(response.Body.get("executionId"))

            tablefields = ["executionId", "outputLocation"]
            Format.FormatResponse(response, lambda r: r.Body, args.format, None, tablefields)

            if args.wait:
                await executionsession.wait_finished()
        finally:
            await executionsession.wait(0)

if __name__ == '__main__':
    if platform.system() == 'Windows':
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
    asyncio.run(main())



# Generated using openapi2py.py
# openapi2py.py "--dhiservice" "sharing" "--classname" "SharingGenClientV" "-r" "projectid" "-i" "datasetid=id" "--out" "C:\work\devops\mike-platform-sdk-py\src\dhi\platform\generated\sharinggen.py" "https://apispec-mike-platform-dev0.eu.mike-cloud-dev0.com/sharing/v1"
# 2022-01-13 19:04:23.671606Z
from typing import Any, Dict, List, NewType, Type, TypeVar, Union
from enum import Enum
import attr
from ..base.client import DataContract, PlatformClient, Response


# https://apispec-mike-platform-dev0.eu.mike-cloud-dev0.com/sharing/v1
# DHI Sharing API - Version 1
# API for managing data publications and subscriptions
# 1

class CatalogTypeV1(str, Enum):
    INTRATENANT = "IntraTenant"
    PUBLIC = "Public"
    def __str__(self) -> str:
        return str(self.value)

class DatasetTypeV1(str, Enum):
    FILE = "file"
    MULTIDIMENSIONAL = "multidimensional"
    TIMESERIES = "timeseries"
    GISVECTORDATA = "gisvectordata"
    TILES = "tiles"
    def __str__(self) -> str:
        return str(self.value)

CreateCatalogOutputV1Type = TypeVar("CreateCatalogOutputV1Type", bound="CreateCatalogOutputV1")

@attr.s(auto_attribs=True)
class CreateCatalogOutputV1(DataContract):
    catalogId: str = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = DataContract.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: CreateCatalogOutputV1Type, src_dict: Dict[str, Any]) -> CreateCatalogOutputV1Type:
        obj = CreateCatalogOutputV1()
        obj.load_dict(src_dict)
        return obj

class ComparisonOperatorV1(str, Enum):
    EQUAL = "Equal"
    NOTEQUAL = "NotEqual"
    GREATERTHAN = "GreaterThan"
    LESSTHAN = "LessThan"
    GREATERTHANOREQUAL = "GreaterThanOrEqual"
    LESSTHANOREQUAL = "LessThanOrEqual"
    def __str__(self) -> str:
        return str(self.value)

class SortOrderV1(str, Enum):
    ASC = "Asc"
    DESC = "Desc"
    def __str__(self) -> str:
        return str(self.value)

CreatePublicationOutputV1Type = TypeVar("CreatePublicationOutputV1Type", bound="CreatePublicationOutputV1")

@attr.s(auto_attribs=True)
class CreatePublicationOutputV1(DataContract):
    publicationId: str = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = DataContract.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: CreatePublicationOutputV1Type, src_dict: Dict[str, Any]) -> CreatePublicationOutputV1Type:
        obj = CreatePublicationOutputV1()
        obj.load_dict(src_dict)
        return obj

EditPublicationInputV1Type = TypeVar("EditPublicationInputV1Type", bound="EditPublicationInputV1")

@attr.s(auto_attribs=True)
class EditPublicationInputV1(DataContract):
    id: str = None
    name: str = None
    description: str = None
    discoverable: str = None
    location: str = None
    resourceId: str = None
    metadata: str = None
    properties: str = None
    rowVersion: str = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = DataContract.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: EditPublicationInputV1Type, src_dict: Dict[str, Any]) -> EditPublicationInputV1Type:
        obj = EditPublicationInputV1()
        obj.load_dict(src_dict)
        return obj

CatalogOutputV1Type = TypeVar("CatalogOutputV1Type", bound="CatalogOutputV1")

@attr.s(auto_attribs=True)
class CatalogOutputV1(DataContract):
    id: str = None
    name: str = None
    catalogType: CatalogTypeV1 = None
    createdBy: str = None
    createdAt: str = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = DataContract.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: CatalogOutputV1Type, src_dict: Dict[str, Any]) -> CatalogOutputV1Type:
        obj = CatalogOutputV1()
        obj.load_dict(src_dict)
        return obj

CatalogOutputCursorResponseV1Type = TypeVar("CatalogOutputCursorResponseV1Type", bound="CatalogOutputCursorResponseV1")

@attr.s(auto_attribs=True)
class CatalogOutputCursorResponseV1(DataContract):
    cursor: str = None
    data: List[CatalogOutputV1] = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = DataContract.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: CatalogOutputCursorResponseV1Type, src_dict: Dict[str, Any]) -> CatalogOutputCursorResponseV1Type:
        obj = CatalogOutputCursorResponseV1()
        obj.load_dict(src_dict)
        return obj

class SpatialOperatorV1(str, Enum):
    INTERSECTS = "Intersects"
    WITHIN = "Within"
    def __str__(self) -> str:
        return str(self.value)

ProblemDetailsV1Type = TypeVar("ProblemDetailsV1Type", bound="ProblemDetailsV1")

@attr.s(auto_attribs=True)
class ProblemDetailsV1(DataContract):
    type: str = None
    title: str = None
    status: int = None
    detail: str = None
    instance: str = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = DataContract.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: ProblemDetailsV1Type, src_dict: Dict[str, Any]) -> ProblemDetailsV1Type:
        obj = ProblemDetailsV1()
        obj.load_dict(src_dict)
        return obj

ValidationProblemDetailsV1Type = TypeVar("ValidationProblemDetailsV1Type", bound="ValidationProblemDetailsV1")

@attr.s(auto_attribs=True)
class ValidationProblemDetailsV1(ProblemDetailsV1):
    errors: str = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = ProblemDetailsV1.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: ValidationProblemDetailsV1Type, src_dict: Dict[str, Any]) -> ValidationProblemDetailsV1Type:
        obj = ValidationProblemDetailsV1()
        obj.load_dict(src_dict)
        return obj

class AttributeOperatorV1(str, Enum):
    EQUAL = "Equal"
    NOTEQUAL = "NotEqual"
    GREATERTHAN = "GreaterThan"
    LESSTHAN = "LessThan"
    GREATERTHANOREQUAL = "GreaterThanOrEqual"
    LESSTHANOREQUAL = "LessThanOrEqual"
    CONTAINS = "Contains"
    STARTSWITH = "StartsWith"
    def __str__(self) -> str:
        return str(self.value)

CreateSubscriptionOutputV1Type = TypeVar("CreateSubscriptionOutputV1Type", bound="CreateSubscriptionOutputV1")

@attr.s(auto_attribs=True)
class CreateSubscriptionOutputV1(DataContract):
    subscriptionId: str = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = DataContract.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: CreateSubscriptionOutputV1Type, src_dict: Dict[str, Any]) -> CreateSubscriptionOutputV1Type:
        obj = CreateSubscriptionOutputV1()
        obj.load_dict(src_dict)
        return obj

CreateCatalogInputV1Type = TypeVar("CreateCatalogInputV1Type", bound="CreateCatalogInputV1")

@attr.s(auto_attribs=True)
class CreateCatalogInputV1(DataContract):
    name: str = None
    catalogType: CatalogTypeV1 = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = DataContract.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: CreateCatalogInputV1Type, src_dict: Dict[str, Any]) -> CreateCatalogInputV1Type:
        obj = CreateCatalogInputV1()
        obj.load_dict(src_dict)
        return obj

class ResourceTypeV1(str, Enum):
    DATASET = "Dataset"
    FOLDER = "Folder"
    def __str__(self) -> str:
        return str(self.value)

ResourceDetailOutputV1Type = TypeVar("ResourceDetailOutputV1Type", bound="ResourceDetailOutputV1")

@attr.s(auto_attribs=True)
class ResourceDetailOutputV1(DataContract):
    customerId: str = None
    resourceId: str = None
    resourceType: ResourceTypeV1 = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = DataContract.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: ResourceDetailOutputV1Type, src_dict: Dict[str, Any]) -> ResourceDetailOutputV1Type:
        obj = ResourceDetailOutputV1()
        obj.load_dict(src_dict)
        return obj

CreateSubscriptionInputV1Type = TypeVar("CreateSubscriptionInputV1Type", bound="CreateSubscriptionInputV1")

@attr.s(auto_attribs=True)
class CreateSubscriptionInputV1(DataContract):
    reference: str = None
    publicationId: str = None
    endDate: str = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = DataContract.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: CreateSubscriptionInputV1Type, src_dict: Dict[str, Any]) -> CreateSubscriptionInputV1Type:
        obj = CreateSubscriptionInputV1()
        obj.load_dict(src_dict)
        return obj

class CatalogSortColumnV1(str, Enum):
    NAME = "Name"
    CREATEDAT = "CreatedAt"
    def __str__(self) -> str:
        return str(self.value)

class SearchSortColumnV1(str, Enum):
    NAME = "Name"
    CREATEDAT = "CreatedAt"
    def __str__(self) -> str:
        return str(self.value)

class PublicationSortColumnV1(str, Enum):
    NAME = "Name"
    CREATEDAT = "CreatedAt"
    def __str__(self) -> str:
        return str(self.value)

QueryConditionV1Type = TypeVar("QueryConditionV1Type", bound="QueryConditionV1")

@attr.s(auto_attribs=True)
class QueryConditionV1(DataContract):
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = DataContract.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: QueryConditionV1Type, src_dict: Dict[str, Any]) -> QueryConditionV1Type:
        obj = QueryConditionV1()
        obj.load_dict(src_dict)
        return obj

FullTextQueryConditionV1Type = TypeVar("FullTextQueryConditionV1Type", bound="FullTextQueryConditionV1")

@attr.s(auto_attribs=True)
class FullTextQueryConditionV1(QueryConditionV1):
    searchString: str = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = QueryConditionV1.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: FullTextQueryConditionV1Type, src_dict: Dict[str, Any]) -> FullTextQueryConditionV1Type:
        obj = FullTextQueryConditionV1()
        obj.load_dict(src_dict)
        return obj

DatasetTypeQueryConditionV1Type = TypeVar("DatasetTypeQueryConditionV1Type", bound="DatasetTypeQueryConditionV1")

@attr.s(auto_attribs=True)
class DatasetTypeQueryConditionV1(QueryConditionV1):
    datasetType: DatasetTypeV1 = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = QueryConditionV1.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: DatasetTypeQueryConditionV1Type, src_dict: Dict[str, Any]) -> DatasetTypeQueryConditionV1Type:
        obj = DatasetTypeQueryConditionV1()
        obj.load_dict(src_dict)
        return obj

SpatialQueryConditionV1Type = TypeVar("SpatialQueryConditionV1Type", bound="SpatialQueryConditionV1")

@attr.s(auto_attribs=True)
class SpatialQueryConditionV1(QueryConditionV1):
    geometry: str = None
    operator: SpatialOperatorV1 = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = QueryConditionV1.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: SpatialQueryConditionV1Type, src_dict: Dict[str, Any]) -> SpatialQueryConditionV1Type:
        obj = SpatialQueryConditionV1()
        obj.load_dict(src_dict)
        return obj

CreatedAtQueryConditionV1Type = TypeVar("CreatedAtQueryConditionV1Type", bound="CreatedAtQueryConditionV1")

@attr.s(auto_attribs=True)
class CreatedAtQueryConditionV1(QueryConditionV1):
    createdAt: str = None
    operator: ComparisonOperatorV1 = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = QueryConditionV1.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: CreatedAtQueryConditionV1Type, src_dict: Dict[str, Any]) -> CreatedAtQueryConditionV1Type:
        obj = CreatedAtQueryConditionV1()
        obj.load_dict(src_dict)
        return obj

IdsQueryConditionV1Type = TypeVar("IdsQueryConditionV1Type", bound="IdsQueryConditionV1")

@attr.s(auto_attribs=True)
class IdsQueryConditionV1(QueryConditionV1):
    ids: List[int] = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = QueryConditionV1.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: IdsQueryConditionV1Type, src_dict: Dict[str, Any]) -> IdsQueryConditionV1Type:
        obj = IdsQueryConditionV1()
        obj.load_dict(src_dict)
        return obj

CatalogIdQueryConditionV1Type = TypeVar("CatalogIdQueryConditionV1Type", bound="CatalogIdQueryConditionV1")

@attr.s(auto_attribs=True)
class CatalogIdQueryConditionV1(QueryConditionV1):
    catalogId: str = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = QueryConditionV1.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: CatalogIdQueryConditionV1Type, src_dict: Dict[str, Any]) -> CatalogIdQueryConditionV1Type:
        obj = CatalogIdQueryConditionV1()
        obj.load_dict(src_dict)
        return obj

SearchInputV1Type = TypeVar("SearchInputV1Type", bound="SearchInputV1")

@attr.s(auto_attribs=True)
class SearchInputV1(DataContract):
    query: List[QueryConditionV1] = None
    sortBy: SearchSortColumnV1 = None
    sortOrder: SortOrderV1 = None
    cursor: str = None
    limit: int = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = DataContract.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: SearchInputV1Type, src_dict: Dict[str, Any]) -> SearchInputV1Type:
        obj = SearchInputV1()
        obj.load_dict(src_dict)
        return obj

class SubscriptionModeV1(str, Enum):
    OPEN = "Open"
    ONAPPROVAL = "OnApproval"
    def __str__(self) -> str:
        return str(self.value)

PublicationOutputV1Type = TypeVar("PublicationOutputV1Type", bound="PublicationOutputV1")

@attr.s(auto_attribs=True)
class PublicationOutputV1(DataContract):
    id: str = None
    catalogId: str = None
    name: str = None
    description: str = None
    location: str = None
    metadata: str = None
    properties: str = None
    resourceId: str = None
    resourceType: ResourceTypeV1 = None
    discoverable: str = None
    subscriptionMode: SubscriptionModeV1 = None
    publishedUntil: str = None
    createdBy: str = None
    createdAt: str = None
    rowVersion: str = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = DataContract.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: PublicationOutputV1Type, src_dict: Dict[str, Any]) -> PublicationOutputV1Type:
        obj = PublicationOutputV1()
        obj.load_dict(src_dict)
        return obj

PublicationOutputCursorResponseV1Type = TypeVar("PublicationOutputCursorResponseV1Type", bound="PublicationOutputCursorResponseV1")

@attr.s(auto_attribs=True)
class PublicationOutputCursorResponseV1(DataContract):
    cursor: str = None
    data: List[PublicationOutputV1] = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = DataContract.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: PublicationOutputCursorResponseV1Type, src_dict: Dict[str, Any]) -> PublicationOutputCursorResponseV1Type:
        obj = PublicationOutputCursorResponseV1()
        obj.load_dict(src_dict)
        return obj

CreatePublicationInputV1Type = TypeVar("CreatePublicationInputV1Type", bound="CreatePublicationInputV1")

@attr.s(auto_attribs=True)
class CreatePublicationInputV1(DataContract):
    name: str = None
    description: str = None
    location: str = None
    metadata: str = None
    properties: str = None
    resourceId: str = None
    catalogId: str = None
    resourceType: ResourceTypeV1 = None
    discoverable: str = None
    subscriptionMode: SubscriptionModeV1 = None
    publishedUntil: str = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = DataContract.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: CreatePublicationInputV1Type, src_dict: Dict[str, Any]) -> CreatePublicationInputV1Type:
        obj = CreatePublicationInputV1()
        obj.load_dict(src_dict)
        return obj

class SubscriptionStateV1(str, Enum):
    APPROVALPENDING = "ApprovalPending"
    APPROVED = "Approved"
    REVOKED = "Revoked"
    def __str__(self) -> str:
        return str(self.value)

SubscriptionOutputV1Type = TypeVar("SubscriptionOutputV1Type", bound="SubscriptionOutputV1")

@attr.s(auto_attribs=True)
class SubscriptionOutputV1(DataContract):
    id: str = None
    reference: str = None
    publicationId: str = None
    publicationName: str = None
    endDate: str = None
    subscriptionState: SubscriptionStateV1 = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = DataContract.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: SubscriptionOutputV1Type, src_dict: Dict[str, Any]) -> SubscriptionOutputV1Type:
        obj = SubscriptionOutputV1()
        obj.load_dict(src_dict)
        return obj

SubscriptionOutputCursorResponseV1Type = TypeVar("SubscriptionOutputCursorResponseV1Type", bound="SubscriptionOutputCursorResponseV1")

@attr.s(auto_attribs=True)
class SubscriptionOutputCursorResponseV1(DataContract):
    cursor: str = None
    data: List[SubscriptionOutputV1] = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = DataContract.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: SubscriptionOutputCursorResponseV1Type, src_dict: Dict[str, Any]) -> SubscriptionOutputCursorResponseV1Type:
        obj = SubscriptionOutputCursorResponseV1()
        obj.load_dict(src_dict)
        return obj

AttributeQueryConditionV1Type = TypeVar("AttributeQueryConditionV1Type", bound="AttributeQueryConditionV1")

@attr.s(auto_attribs=True)
class AttributeQueryConditionV1(QueryConditionV1):
    name: str = None
    operator: AttributeOperatorV1 = None
    value: None = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = QueryConditionV1.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: AttributeQueryConditionV1Type, src_dict: Dict[str, Any]) -> AttributeQueryConditionV1Type:
        obj = AttributeQueryConditionV1()
        obj.load_dict(src_dict)
        return obj

SearchOutputV1Type = TypeVar("SearchOutputV1Type", bound="SearchOutputV1")

@attr.s(auto_attribs=True)
class SearchOutputV1(DataContract):
    id: str = None
    name: str = None
    location: str = None
    resourceId: str = None
    resourceType: ResourceTypeV1 = None
    catalogId: str = None
    catalogName: str = None
    subscriptionMode: SubscriptionModeV1 = None
    publishedUntil: str = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = DataContract.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: SearchOutputV1Type, src_dict: Dict[str, Any]) -> SearchOutputV1Type:
        obj = SearchOutputV1()
        obj.load_dict(src_dict)
        return obj

SearchOutputCursorResponseV1Type = TypeVar("SearchOutputCursorResponseV1Type", bound="SearchOutputCursorResponseV1")

@attr.s(auto_attribs=True)
class SearchOutputCursorResponseV1(DataContract):
    cursor: str = None
    data: List[SearchOutputV1] = None
    __renamed = {  }
    def to_dict(self) -> Dict[str, Any]:
        return self.get_dictionary(self.get_renamed())
    def load_dict(self, src_dict: Dict[str, Any]) -> None:
        DataContract.load_from_directory(self, src_dict, self.get_renamed())
    @classmethod
    def get_renamed(cls) -> Dict[str, str]:
        result = DataContract.get_renamed().copy()
        result.update(cls.__renamed)
        return result
    @classmethod
    def from_dict(cls: SearchOutputCursorResponseV1Type, src_dict: Dict[str, Any]) -> SearchOutputCursorResponseV1Type:
        obj = SearchOutputCursorResponseV1()
        obj.load_dict(src_dict)
        return obj

class SubscriptionSortColumnV1(str, Enum):
    CREATEDAT = "CreatedAt"
    def __str__(self) -> str:
        return str(self.value)

class SharingGenClientV1(PlatformClient):
    def __init__(self, inspectFnc=PlatformClient.DefaultInspectFnc, **kwargs):
        super().__init__(inspectFnc, includeheaders=PlatformClient.GetServiceHeaders("sharing"), **kwargs)

    def CreateCatalog(self, projectid, body) -> Response:
        """Catalog

        POST /api/sharing/catalog
        """
        return self.PostRequest("/api/sharing/catalog", body, None, api_version="1", projectid=projectid)

    def GetCatalogList(self, projectid, sortby=None, sortorder=None, cursor=None, limit=None) -> Response:
        """Catalog

        GET /api/sharing/catalog/list
        """
        queryparams = self.GetQueryParams(SortBy=sortby, SortOrder=sortorder, Cursor=cursor, Limit=limit)
        return self.GetRequest("/api/sharing/catalog/list", queryparams, api_version="1", projectid=projectid)

    def CreatePublication(self, projectid, body) -> Response:
        """Publication

        POST /api/sharing/publication
        """
        return self.PostRequest("/api/sharing/publication", body, None, api_version="1", projectid=projectid)

    def UpdatePublication(self, projectid, body) -> Response:
        """Publication

        PUT /api/sharing/publication
        """
        return self.PutRequest("/api/sharing/publication", body, None, api_version="1", projectid=projectid)

    def GetPublicationList(self, projectid, resourceid=None, catalogid=None, sortby=None, sortorder=None, cursor=None, limit=None) -> Response:
        """Publication

        GET /api/sharing/publication/list
        """
        queryparams = self.GetQueryParams(ResourceId=resourceid, CatalogId=catalogid, SortBy=sortby, SortOrder=sortorder, Cursor=cursor, Limit=limit)
        return self.GetRequest("/api/sharing/publication/list", queryparams, api_version="1", projectid=projectid)

    def DeletePublication(self, projectid, id) -> Response:
        """Publication

        DELETE /api/sharing/publication/{id}
        """
        return self.DeleteRequest(f"/api/sharing/publication/{id}", None, api_version="1", projectid=projectid, datasetid=id)

    def GetPublication(self, projectid, id) -> Response:
        """Publication

        GET /api/sharing/publication/{id}
        """
        return self.GetRequest(f"/api/sharing/publication/{id}", None, api_version="1", projectid=projectid, datasetid=id)

    def GetPublicationSubscriptionList(self, projectid, id, sortby=None, sortorder=None, cursor=None, limit=None) -> Response:
        """Publication

        GET /api/sharing/publication/{id}/subscription/list
        """
        queryparams = self.GetQueryParams(SortBy=sortby, SortOrder=sortorder, Cursor=cursor, Limit=limit)
        return self.GetRequest(f"/api/sharing/publication/{id}/subscription/list", queryparams, api_version="1", projectid=projectid, datasetid=id)

    def GetResourceDetail(self, projectid, subscriptionid) -> Response:
        """Publication

        GET /api/sharing/publication/{subscriptionId}/resourcedetail
        """
        return self.GetRequest(f"/api/sharing/publication/{subscriptionid}/resourcedetail", None, api_version="1", projectid=projectid)

    def Search(self, projectid, body) -> Response:
        """Search

        POST /api/sharing/search
        """
        return self.PostRequest("/api/sharing/search", body, None, api_version="1", projectid=projectid)

    def CreateSubscription(self, projectid, body) -> Response:
        """Subscription

        POST /api/sharing/subscription
        """
        return self.PostRequest("/api/sharing/subscription", body, None, api_version="1", projectid=projectid)

    def GetSubscriptionList(self, projectid, cursor=None, limit=None) -> Response:
        """Subscription

        GET /api/sharing/subscription/list
        """
        queryparams = self.GetQueryParams(Cursor=cursor, Limit=limit)
        return self.GetRequest("/api/sharing/subscription/list", queryparams, api_version="1", projectid=projectid)



#!/usr/bin/env python
from dhi.platform.args import ClientArgs
from dhi.platform.config import ClientConfig
from dhi.platform.timeserieshelper import TimeSeriesV2Contracts
from dhi.platform.generated.timeseriesgen import TimeSeriesGenClientV2
from dhi.platform.fmt import Format

def __initParser(parser):
    parser.add_argument("-f", "--inputfile", help="Input file path, '-' to read stdin")
    parser.add_argument("--schemaproperties", help="Time series schema properties")
    parser.add_argument("--name", help="Time series name")
    parser.add_argument("--description", help="Time series description")
    parser.add_argument("--metadata", help="Time series metadata JSON, {}")
    parser.add_argument("--properties", help="Time series settings JSON, {}")

def main():
    args = ClientArgs.ParseForProject(description="Create time series", init=__initParser, defaultformat="yaml")
    ClientConfig.UpdateProjectFromConfiguration(args)
    clientv2 = TimeSeriesGenClientV2(**vars(args))

    input = TimeSeriesV2Contracts.PrepareCreateTimeseriesDatasetV2Input(
        ClientArgs.LoadJson(args.inputfile),
        schemaproperties=args.schemaproperties,
        name=args.name,
        description=args.description,
        metadata=ClientArgs.LoadJsonStr(args.metadata),
        properties=ClientArgs.LoadJsonStr(args.properties))
    response = clientv2.CreateTimeseriesDatasetV2(args.projectid, input)

    tablefields = ["id", "name"]
    Format.FormatResponse(response, lambda r: r.Body, args.format, None, tablefields)

if __name__ == '__main__':
    main()
